# Module for flux calibrating spectra
import numpy as np
import os
import matplotlib.pyplot as plt
from astropy.io import fits

from pypeit import msgs
from pypeit.spectrographs.util import load_spectrograph
from pypeit import sensfunc
from pypeit import specobjs
from astropy import table
from IPython import embed



class FluxCalibrate(object):
    """
    Class for flux calibrating spectra.

    Args:
        spec1dfiles (list):
        sensfiles (list):
        par (pypeit.par.pypeitpar.FluxCalibrate, optional):
        outfiles (list, optional):
            Names of the outfiles.  If None, this is set to spec1dfiles and those are overwritten
    """
    # Superclass factory method generates the subclass instance
    @classmethod
    def get_instance(cls, spec1dfiles, sensfiles, par=None, debug=False):
        pypeline = fits.getheader(spec1dfiles[0])['PYPELINE'] + 'FC'
        return next(c for c in cls.__subclasses__() if c.__name__ == pypeline)(
            spec1dfiles, sensfiles, par=par, debug=debug)

    def __init__(self, spec1dfiles, sensfiles, par=None, debug=False, outfiles=None):

        self.spec1dfiles = spec1dfiles
        self.sensfiles = sensfiles

        # Output file names
        self.outfiles = spec1dfiles if outfiles is None else outfiles

        # Load the spectrograph
        header = fits.getheader(spec1dfiles[0])
        self.spectrograph = load_spectrograph(header['PYP_SPEC'])
        self.par = self.spectrograph.default_pypeit_par()['fluxcalib'] if par is None else par
        self.debug = debug

        sens_last = None
        for spec1, sens, outfile in zip(self.spec1dfiles, self.sensfiles, self.outfiles):
            # Read in the data
            sobjs = specobjs.SpecObjs.from_fitsfile(spec1)
            if sens != sens_last:
                wave, sensfunction, meta_table, out_table, header_sens = sensfunc.SensFunc.load(sens)
            self.flux_calib(sobjs, wave, sensfunction, meta_table)
            sobjs.write_to_fits(sobjs.header, outfile, overwrite=True)

    def flux_calib(self, sobjs, wave, sensfunction, meta_table):
        """
        Dummy method overloaded by subclass

        Args:
            sobjs:
            wave:
            sensfunction:
            meta_table:


        """
        pass

class MultiSlitFC(FluxCalibrate):
    """
    Child of FluxSpec for Multislit and Longslit reductions
    """

    def __init__(self, spec1dfiles, sensfiles, par=None, debug=False, outfiles=None):
        super().__init__(spec1dfiles, sensfiles, par=par, debug=debug, outfiles=outfiles)


    def flux_calib(self, sobjs, wave, sensfunction, meta_table):
        """
        Apply sensitivity function to all the spectra in an sobjs object.

        Args:
            sobjs (object):
               SpecObjs object
            wave (ndarray):
               wavelength array for sensitivity function (nspec,)
            sensfunction (ndarray):
               sensitivity function
            meta_table (table):
               astropy table containing meta data for sensitivity function

        """

        # Run
        for sci_obj in sobjs:
            sci_obj.apply_flux_calib(wave, sensfunction,
                                     sobjs.header['EXPTIME'],
                                     extinct_correct=self.par['extinct_correct'],
                                     longitude=self.spectrograph.telescope['longitude'],
                                     latitude=self.spectrograph.telescope['latitude'],
                                     airmass=float(sobjs.header['AIRMASS']))




class EchelleFC(FluxCalibrate):
    """
    Child of FluxSpec for Echelle reductions
    """

    def __init__(self, spec1dfiles, sensfiles, par=None, debug=False):
        super().__init__(spec1dfiles, sensfiles, par=par, debug=debug)


    def flux_calib(self, sobjs, wave, sensfunction, meta_table):
        """
        Apply sensitivity function to all the spectra in an sobjs object.

        Args:
            sobjs (object):
               SpecObjs object
            wave (ndarray):
               wavelength array for sensitivity function (nspec,)
            sensfunction (ndarray):
               sensitivity function
            meta_table (table):
               astropy table containing meta data for sensitivity function

        """

        # Flux calibrate the orders that are mutually in the meta_table and in the sobjs. This allows flexibility
        # for applying to data for cases where not all orders are present in the data as in the sensfunc, etc.,
        # i.e. X-shooter with the K-band blocking filter.
        ech_orders = np.array(meta_table['ECH_ORDERS']).flatten()
        #norders = ech_orders.size
        for sci_obj in sobjs:
            # JFH Is there a more elegant pythonic way to do this without looping over both orders and sci_obj?
            indx = np.where(ech_orders == sci_obj.ECH_ORDER)[0]
            if indx.size==1:
                sci_obj.apply_flux_calib(wave[:, indx[0]],sensfunction[:,indx[0]],
                                         sobjs.header['EXPTIME'],
                                         extinct_correct=self.par['extinct_correct'],
                                         longitude=self.spectrograph.telescope['longitude'],
                                         latitude=self.spectrograph.telescope['latitude'],
                                         airmass=float(sobjs.header['AIRMASS']))
            elif indx.size == 0:
                msgs.info('Unable to flux calibrate order = {:} as it is not in your sensitivity function. '
                          'Something is probably wrong with your sensitivity function.'.format(sci_obj.ECH_ORDER))
            else:
                msgs.error('This should not happen')



"""
Module for guiding construction of the Wavelength Image

.. include common links, assuming primary doc root is up one directory
.. include:: ../links.rst
"""
import inspect

import numpy as np
import os

from pypeit import msgs
from pypeit import utils
from pypeit import masterframe
from pypeit import edgetrace
from pypeit.core import pixels
from pypeit.core import save
from pypeit.core import load
from IPython import embed


class WaveImage(masterframe.MasterFrame):
    """
    Class to generate the Wavelength Image

    Args:
        tslits_dict (dict or None):
            dict from TraceSlits class (e.g. slitpix)
        tilts (np.ndarray or None):
            Tilt image
        wv_calib (dict or None): wavelength solution dictionary
            Parameters are read from wv_calib['par']
        spectrograph (:class:`pypeit.spectrographs.spectrograph.Spectrograph`):
            The `Spectrograph` instance that sets the
            instrument used to take the observations.  Used to set
            :attr:`spectrograph`.
        det (int or None):
        maskslits (np.ndarray or None):
            True = skip this slit
        master_key (:obj:`str`, optional):
            The string identifier for the instrument configuration.  See
            :class:`pypeit.masterframe.MasterFrame`.
        master_dir (str, optional):
            Path to master frames
        reuse_masters (bool, optional):
            Load from disk if possible

    Attributes:
        image (np.ndarray): Wavelength image
        steps (list): List of the processing steps performed

    """
    master_type = 'Wave'

    @classmethod
    def from_master_file(cls, master_file):
        """

        Args:
            master_file (str):

        Returns:
            waveimage.WaveImage:

        """
        # Spectrograph
        spectrograph, extras = masterframe.items_from_master_file(master_file)
        head0 = extras[0]
        # Master info
        master_dir = head0['MSTRDIR']
        master_key = head0['MSTRKEY']
        # Instantiate
        slf = cls(None, None, None, spectrograph, None, None, master_dir=master_dir, master_key=master_key,
                  reuse_masters=True)
        slf.image = slf.load(ifile=master_file)
        # Return
        return slf

    def __init__(self, tslits_dict, tilts, wv_calib, spectrograph, det, maskslits,
                 master_key=None, master_dir=None, reuse_masters=False):


        # MasterFrame
        masterframe.MasterFrame.__init__(self, self.master_type, master_dir=master_dir,
                                         master_key=master_key, reuse_masters=reuse_masters)
        # Required parameters
        self.spectrograph = spectrograph
        self.det = det

        self.tslits_dict = tslits_dict
        self.tilts = tilts
        self.wv_calib = wv_calib
        if tslits_dict is not None:
            self.slitmask = pixels.tslits2mask(self.tslits_dict)
            self.slit_spat_pos = edgetrace.slit_spat_pos(self.tslits_dict['slit_left'],
                                                         self.tslits_dict['slit_righ'],
                                                         self.tslits_dict['nspec'],
                                                         self.tslits_dict['nspat'])
        else:
            self.slitmask = None
            self.slit_spat_pos = None

        self.maskslits = maskslits

        # For echelle order, primarily
        # TODO: only echelle is ever used.  Do we need to keep the whole
        # thing?
        self.par = wv_calib['par'] if wv_calib is not None else None

        # Main output
        self.image = None
        self.steps = []

    def build_wave(self):
        """
        Main algorithm to build the wavelength image

        Returns:
            `numpy.ndarray`_: The wavelength image.

        """
        # Loop on slits
        ok_slits = np.where(np.invert(self.maskslits))[0]
        self.image = np.zeros_like(self.tilts)
        nspec = self.slitmask.shape[0]

        # Error checking on the wv_calib
        #if (nspec-1) != int(self.wv_calib[str(0)]['fmax']):
        #    msgs.error('Your wavelength fits used inconsistent normalization. Something is wrong!')

        # If this is echelle print out a status message and do some error checking
        if self.par['echelle']:
            msgs.info('Evaluating 2-d wavelength solution for echelle....')
            if len(self.wv_calib['fit2d']['orders']) != len(ok_slits):
                msgs.error('wv_calib and ok_slits do not line up. Something is very wrong!')

        # Unpack some 2-d fit parameters if this is echelle
        for slit in ok_slits:
            thismask = (self.slitmask == slit)
            if self.par['echelle']:
                order, indx = self.spectrograph.slit2order(self.slit_spat_pos[slit])
                # evaluate solution
                self.image[thismask] = utils.func_val(self.wv_calib['fit2d']['coeffs'],
                                                       self.tilts[thismask],
                                                       self.wv_calib['fit2d']['func2d'],
                                                       x2=np.ones_like(self.tilts[thismask])*order,
                                                       minx=self.wv_calib['fit2d']['min_spec'],
                                                       maxx=self.wv_calib['fit2d']['max_spec'],
                                                       minx2=self.wv_calib['fit2d']['min_order'],
                                                       maxx2=self.wv_calib['fit2d']['max_order'])
                self.image[thismask] /= order
            else:
                iwv_calib = self.wv_calib[str(slit)]
                self.image[thismask] = utils.func_val(iwv_calib['fitc'], self.tilts[thismask],
                                                       iwv_calib['function'],
                                                       minx=iwv_calib['fmin'],
                                                       maxx=iwv_calib['fmax'])

        # Return
        self.steps.append(inspect.stack()[0][3])
        return self.image

    def __repr__(self):
        # Generate sets string
        txt = '<{:s}: >'.format(self.__class__.__name__)
        return txt

    def save(self, outfile=None, overwrite=True, image=None):
        """
        Save the master wavelength image.

        Args:
            outfile (:obj:`str`, optional):
                Name for the output file.  Defaults to
                :attr:`file_path`.
            overwrite (:obj:`bool`, optional):
                Overwrite any existing file.
        """
        _outfile = self.master_file_path if outfile is None else outfile
        # Check if it exists
        if os.path.exists(_outfile) and not overwrite:
            msgs.warn('Master file exists: {0}'.format(_outfile) + msgs.newline()
                      + 'Set overwrite=True to overwrite it.')
            return
        # Setup the items
        hdr = self.build_master_header(steps=self.steps)
        _image = self.image if image is None else image
        # Save to a multi-extension FITS
        save.write_fits(hdr, [_image], _outfile, extnames=['WAVE'])
        msgs.info('Master frame written to {0}'.format(_outfile))

    def load(self, ifile=None, return_header=False):
        """
        Load the wavelength image data from a saved master frame.

        Args:
            ifile (:obj:`str`, optional):
                Name of the master frame file.  Defaults to
                :attr:`file_path`.
            return_header (:obj:`bool`, optional):
                Return the header

        Returns:
            tuple: Returns an `numpy.ndarray`_ with the wavelength image.
        """
        #return super(WaveImage, self).load('WAVE', ifile=ifile, return_header=return_header)
        master_file = self.chk_load_master(ifile)
        if master_file is None:
            return None
        # Load
        self.image, head0 = load.load_multiext_fits(master_file, ['WAVE'])
        # Return
        return self.image

    '''
    @staticmethod
    def load_from_file(filename, return_header=False):
        """
        Load the wavelength image data from a saved master frame.

        Args:
            filename (:obj:`str`, optional):
                Name of the master frame file.
            return_header (:obj:`bool`, optional):
                Return the header

        Returns:
            tuple: Returns an `numpy.ndarray`_ with the wavelength
            image.  Also returns the primary header, if requested.
        """
        # Use of super() to call staticmethods of the base class seems
        # like a bit of a mess (bound vs. unbound methods).  There's a
        # syntax that works, but for now, I'm just going to call the
        # static method explicitly without using super().  See:
        # https://stackoverflow.com/questions/26788214/super-and-staticmethod-interaction
        return MasterFrame.load_from_file(filename, 'WAVE', return_header=return_header)
    '''




# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.

import sys
import os
import glob

from setuptools import setup, find_packages


def get_data_files():
    """ Build the list of data files to include.  """
    data_files = []

    # Walk through the data directory, adding all files
    data_generator = os.walk('pypeit/data')
    for path, directories, files in data_generator:
        for f in files:
            data_path = '/'.join(path.split('/')[1:])
            data_files.append(os.path.join(data_path, f))

    # Add pipeline and spectrograph settings
    settings = glob.glob('pypeit/settings/settings.*')
    settings = ['/'.join(path.split('/')[1:]) for path in settings]
    data_files.extend(settings)

    return data_files


def get_scripts():
    """ Grab all the scripts in the bin directory.  """
    scripts = []
    if os.path.isdir('bin'):
        scripts = [ fname for fname in glob.glob(os.path.join('bin', '*'))
                                if not os.path.basename(fname).endswith('.rst') ]
    return scripts


def get_requirements():
    """ Get the requirements from a system file.  """
    name = 'pypeit/requirements.txt'

    requirements_file = os.path.join(os.path.dirname(__file__), name)
    install_requires = [line.strip().replace('==', '>=') for line in open(requirements_file)
                        if not line.strip().startswith('#') and line.strip() != '']
    return install_requires


NAME = 'pypeit'
# do not use x.x.x-dev.  things complain.  instead use x.x.xdev
VERSION = '0.13.1'
RELEASE = 'dev' not in VERSION

def run_setup(data_files, scripts, packages, install_requires):

    # TODO: Are any/all of the *'d keyword arguments needed? I.e., what
    # are the default values?

    setup(name=NAME,
          provides=NAME,                                                # *
          version=VERSION,
          license='BSD3',
          description='PypeIt Spectroscopic Reduction',
          long_description=open('README.md').read(),
          author='PypeIt Collaboration',
          author_email='pypeit@ucolick.org',
          keywords='pypeit PypeIt astronomy Keck UCO Lick data reduction',
          url='https://github.com/pypeit/PypeIt',
          packages=packages,
          package_data={'pypeit': data_files, '': ['*.rst', '*.txt']},
          include_package_data=True,
          scripts=scripts,
          install_requires=install_requires,
          requires=[ 'Python (>3.6.0)' ],                               # *
          zip_safe=False,                                               # *
          use_2to3=False,                                               # *
          setup_requires=[ 'pytest-runner' ],
          tests_require=[ 'pytest' ],
          classifiers=[
              'Development Status :: 4 - Beta',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: BSD License',
              'Natural Language :: English',
              'Operating System :: OS Independent',
              'Programming Language :: Python',
              'Programming Language :: Python :: 3.6',
              'Topic :: Documentation :: Sphinx',
              'Topic :: Scientific/Engineering :: Astronomy',
              'Topic :: Software Development :: Libraries :: Python Modules',
              'Topic :: Software Development :: User Interfaces'
          ],
          )

#-----------------------------------------------------------------------
if __name__ == '__main__':

    # Compile the data files to include
    data_files = get_data_files()
    # Compile the scripts in the bin/ directory
    scripts = get_scripts()
    # Get the packages to include
    packages = find_packages()
    # Collate the dependencies based on the system text file
    install_requires = get_requirements()
    install_requires = []  # Remove this line to enforce actual installation
    # Run setup from setuptools
    run_setup(data_files, scripts, packages, install_requires)





"""
Module to run tests loading
"""
import os
import pytest
from pypeit import specobjs
from pypeit import specobj

def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)



def test_load_specobjs():
    spec_file = data_path('spec1d_r153-J0025-0312_KASTr_2015Jan23T025323.850.fits')
    sobjs = specobjs.SpecObjs.from_fitsfile(spec_file)

    # Test
    assert isinstance(sobjs, specobjs.SpecObjs)
    assert len(sobjs[0].BOX_COUNTS) == 1200

    assert isinstance(sobjs[0], specobj.SpecObj)




"""
Module to run tests on scripts
"""
import os
import glob
import shutil

import numpy as np

import pytest
import yaml

from pypeit.pypmsgs import PypeItError
from pypeit.par.util import parse_pypeit_file
from pypeit.scripts import setup
from pypeit.tests.tstutils import dev_suite_required


def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)


expected = ['pypeit', 'sorted']

def test_run_setup():
    """ Test the setup script
    """
    # Remove .setup if needed
    sfiles = glob.glob('*.setups')
    for sfile in sfiles:
        os.remove(sfile)
    #
    droot = data_path('b')
    pargs = setup.parser(['-r', droot, '-s', 'shane_kast_blue', '-c=all',
                          '--extension=fits.gz', '--output_path={:s}'.format(data_path(''))])
    setup.main(pargs)

    '''
    setup_file = glob.glob(data_path('setup_files/shane_kast_blue*.setups'))[0]
    # Load
    with open(setup_file, 'r') as infile:
        setup_dict = yaml.load(infile)
    # Test
    assert '01' in setup_dict['A'].keys()
    assert setup_dict['A']['--']['disperser']['name'] == '600/4310'
    '''
    # Failures
    pargs2 = setup.parser(['-r', droot, '-s', 'shane_kast_blu', '-c=all',
                              '--extension=fits.gz', '--output_path={:s}'.format(data_path(''))])
    with pytest.raises(ValueError):
        setup.main(pargs2)


def test_setup_made_pypeit_file():
    """ Test the .pypeit file(s) made by pypeit_setup
    This test depends on the one above
    """
    pypeit_file = data_path('shane_kast_blue_A/shane_kast_blue_A.pypeit')
    cfg_lines, data_files, frametype, usrdata, setups = parse_pypeit_file(pypeit_file)
    # Test
    assert len(data_files) == 2
    assert sorted(frametype['b1.fits.gz'].split(',')) == ['arc', 'tilt']
    assert setups[0] == 'A'


@dev_suite_required
def test_setup_keck_lris_red():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/keck_lris_red/multi_400_8500_d560')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'keck_lris_red'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'keck_lris_red*'))
    ext = [f.split('.')[-1] for f in files]
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)


@dev_suite_required
def test_setup_keck_lris_blue():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/keck_lris_blue/multi_600_4000_d560')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'keck_lris_blue'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'keck_lris_blue*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_shane_kast_blue():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/shane_kast_blue/600_4310_d55')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'shane_kast_blue'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'shane_kast_blue*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_shane_kast_red():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/shane_kast_red/600_7500_d55_ret')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'shane_kast_red'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'shane_kast_red*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

# TODO: We need a test data set for shane_kast_red_ret

@dev_suite_required
def test_setup_keck_deimos():

    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/keck_deimos/830G_M_8600')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'keck_deimos'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'keck_deimos*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_keck_nires():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/keck_nires/NIRES/')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'keck_nires'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'keck_nires*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_keck_nirspec():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/keck_nirspec/LOW_NIRSPEC-1')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'keck_nirspec_low'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'keck_nirspec*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)


@dev_suite_required
def test_setup_magellan_mage():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/magellan_mage/1x1')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'magellan_mage'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'magellan_mage*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_wht_isis_blue():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/wht_isis_blue/long_R300B_d5300')
    droot += '/'
    pargs = setup.parser(['-r', droot, '-s', 'wht_isis_blue', '--extension', '.fit'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'wht_isis_blue*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_vlt_xshooter_uvb():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/vlt_xshooter/UVB_1x1')
    droot += '/XSHO'
    pargs = setup.parser(['-r', droot, '-s', 'vlt_xshooter_uvb'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'vlt_xshooter_uvb*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_vlt_xshooter_vis():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/vlt_xshooter/VIS_1x1')
    droot += '/XSHO'
    pargs = setup.parser(['-r', droot, '-s', 'vlt_xshooter_vis'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'vlt_xshooter_vis*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_vlt_xshooter_nir():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/vlt_xshooter/NIR')
    droot += '/XSHO'
    pargs = setup.parser(['-r', droot, '-s', 'vlt_xshooter_nir'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'vlt_xshooter_nir*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

@dev_suite_required
def test_setup_gemini_gnirs():
    droot = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA/gemini_gnirs/32_SB_SXD/')
    droot += '/cN'
    pargs = setup.parser(['-r', droot, '-s', 'gemini_gnirs'])
    setup.main(pargs)

    cwd = os.getcwd()
    setup_dir = os.path.join(cwd, 'setup_files')
    assert os.path.isdir(setup_dir), 'No setup_files directory created'

    files = glob.glob(os.path.join(setup_dir, 'gemini_gnirs*'))
    ext = [f.split('.')[-1] for f in files]
    #expected = ['lst', 'pypeit', 'setups', 'sorted']
    assert np.all([e in ext for e in expected]), \
            'Did not find all setup file extensions: {0}'.format(expected)

    # Clean-up
    shutil.rmtree(setup_dir)

# TODO: Add other instruments!




import numpy as np

#from matplotlib import pyplot as plt
#
## Force the default matplotlib plotting parameters
#plt.rcdefaults()

MAX_REJECT = 0.5
MIN_NPIXELS = 5
GOOD_PIXEL = 0
BAD_PIXEL = 1
KREJ = 2.5
MAX_ITERATIONS = 5


def zscale(image, nsamples=1000, contrast=0.25, bpmask=None, zmask=None):
    """
    Implement IRAF zscale algorithm
    nsamples=1000 and contrast=0.25 are the IRAF display task defaults
    bpmask and zmask not implemented yet
    image is a 2-d np array
    returns (z1, z2)
    """

    # Sample the image
    samples = zsc_sample(image, nsamples, bpmask, zmask)
    npix = len(samples)
    samples.sort()
    zmin = samples[0]
    zmax = samples[-1]
    # For a zero-indexed array
    center_pixel = (npix - 1) // 2
    if npix % 2 == 1:
        median = samples[center_pixel]
    else:
        median = 0.5 * (samples[center_pixel] + samples[center_pixel + 1])

    #
    # Fit a line to the sorted array of samples
    minpix = np.max([MIN_NPIXELS, int(npix * MAX_REJECT)])
    ngrow = np.max([1, int (npix * 0.01)])
    ngoodpix, zstart, zslope = zsc_fit_line(samples, npix, KREJ, ngrow, MAX_ITERATIONS)

    if ngoodpix < minpix:
        z1 = zmin
        z2 = zmax
    else:
        if contrast > 0: zslope /= contrast
        z1 = np.max([zmin, median - (center_pixel - 1) * zslope])
        z2 = np.min([zmax, median + (npix - center_pixel) * zslope])
    return z1, z2


def zsc_sample(image, maxpix, bpmask=None, zmask=None):
    # Figure out which pixels to use for the zscale algorithm
    # Returns the 1-d array samples
    # Don't worry about the bad pixel mask or zmask for the moment
    # Sample in a square grid, and return the first maxpix in the sample
    nc = image.shape[0]
    nl = image.shape[1]
    stride = np.max([1.0, np.sqrt((nc - 1) * (nl - 1) / float(maxpix))])
    stride = int(stride)
    samples = image[::stride,::stride].flatten()
    return samples[:maxpix]


def zsc_fit_line(samples, npix, krej, ngrow, maxiter):
    # First re-map indices from -1.0 to 1.0
    xscale = 2.0 / (npix - 1)
    xnorm = np.arange(npix)
    xnorm = xnorm * xscale - 1.0

    ngoodpix = npix
    minpix = np.max([MIN_NPIXELS, int (npix*MAX_REJECT)])
    last_ngoodpix = npix + 1

    # This is the mask used in k-sigma clipping.0 is good, 1 is bad
    badpix = np.zeros(npix, dtype="int32")

    #Iterate
    for niter in range(maxiter):

        if (ngoodpix >= last_ngoodpix) or (ngoodpix < minpix):
            break

        # Accumulate sums to calculate straight line fit
        goodpixels = np.where(badpix == GOOD_PIXEL)
        sumx = xnorm[goodpixels].sum()
        sumxx = (xnorm[goodpixels]*xnorm[goodpixels]).sum()
        sumxy = (xnorm[goodpixels]*samples[goodpixels]).sum()
        sumy = samples[goodpixels].sum()
        sum = len(goodpixels[0])

        delta = sum * sumxx - sumx * sumx
        # Slope and intercept
        intercept = (sumxx * sumy - sumx * sumxy) / delta
        slope = (sum * sumxy - sumx * sumy) / delta

        # Subtract fitted line from the data array
        fitted = xnorm*slope + intercept
        flat = samples - fitted

        # Compute the k-sigma rejection threshold
        ngoodpix, mean, sigma = zsc_compute_sigma(flat, badpix)

        threshold = sigma * krej

        # Detect and reject pixels further than k*sigma from the fitted line
        lcut = -threshold
        hcut = threshold
        below = np.where(flat < lcut)
        above = np.where(flat > hcut)

        badpix[below] = BAD_PIXEL
        badpix[above] = BAD_PIXEL

        # Convolve with a kernel of length ngrow
        kernel = np.ones(ngrow, dtype="int32")
        badpix = np.convolve(badpix, kernel, mode='same')

        ngoodpix = len(np.where(badpix == GOOD_PIXEL)[0])

        niter += 1

    # Transform the line coefficients back to the X range [0:npix-1]
    zstart = intercept - slope
    zslope = slope * xscale

    return ngoodpix, zstart, zslope


def zsc_compute_sigma(flat, badpix):
    """
    Compute the rms deviation from the mean of a flattened array.
    Ignore rejected pixels
    """

    # Accumulate sum and sum of squares
    goodpixels = np.where(badpix == GOOD_PIXEL)
    sumz = flat[goodpixels].sum()
    sumsq = (flat[goodpixels]*flat[goodpixels]).sum()
    ngoodpix = len(goodpixels[0])
    if ngoodpix == 0:
        mean = None
        sigma = None
    elif ngoodpix == 1:
        mean = sumz
        sigma = None
    else:
        mean = sumz / ngoodpix
        temp = sumsq / (ngoodpix - 1) - sumz*sumz / (ngoodpix * (ngoodpix - 1))
        if temp < 0:
            sigma = 0.0
        else:
            sigma = np.sqrt (temp)

    return ngoodpix, mean, sigma



""" Module for loading PypeIt files
"""
import os
import warnings

import numpy as np

from astropy import units
from astropy.time import Time
from astropy.io import fits
from astropy.table import Table

from linetools.spectra.xspectrum1d import XSpectrum1D
from linetools.spectra.utils import collate
import linetools.utils

from pypeit import msgs
from IPython import embed
from pypeit.core import parse



# TODO I don't think we need this routine
def load_ext_to_array(hdulist, ext_id, ex_value='OPT', flux_value=True, nmaskedge=None):
    '''
    It will be called by load_1dspec_to_array.
    Load one-d spectra from ext_id in the hdulist

    Args:
        hdulist: FITS HDU list
        ext_id: extension name, i.e., 'SPAT1073-SLIT0001-DET03', 'OBJID0001-ORDER0003', 'OBJID0001-ORDER0002-DET01'
        ex_value: 'OPT' or 'BOX'
        flux_value: if True load fluxed data, else load unfluxed data

    Returns:
        tuple: Returns wave, flux, ivar, mask
    '''

    if (ex_value != 'OPT') and (ex_value != 'BOX'):
        msgs.error('{:} is not recognized. Please change to either BOX or OPT.'.format(ex_value))

    # get the order/slit information
    ntrace0 = np.size(hdulist)-1
    idx_names = []
    for ii in range(ntrace0):
        idx_names.append(hdulist[ii+1].name) # idx name

    # Initialize ext
    ext = None
    for indx in (idx_names):
        if ext_id in indx:
            ext = indx
    if ext is None:
        msgs.error('Can not find extension {:}.'.format(ext_id))
    else:
        hdu_iexp = hdulist[ext]

    wave = hdu_iexp.data['{:}_WAVE'.format(ex_value)]
    mask = hdu_iexp.data['{:}_MASK'.format(ex_value)]

    # Mask Edges
    if nmaskedge is not None:
        mask[:int(nmaskedge)] = False
        mask[-int(nmaskedge):] = False

    if flux_value:
        flux = hdu_iexp.data['{:}_FLAM'.format(ex_value)]
        ivar = hdu_iexp.data['{:}_FLAM_IVAR'.format(ex_value)]
    else:
        msgs.warn('Loading unfluxed spectra')
        flux = hdu_iexp.data['{:}_COUNTS'.format(ex_value)]
        ivar = hdu_iexp.data['{:}_COUNTS_IVAR'.format(ex_value)]

    return wave, flux, ivar, mask

# TODO merge this with unpack orders
def load_1dspec_to_array(fnames, gdobj=None, order=None, ex_value='OPT', flux_value=True, nmaskedge=None):
    '''
    Load the spectra from the 1d fits file into arrays.
    If Echelle, you need to specify which order you want to load.
    It can NOT load all orders for Echelle data.

    Args:
        fnames (list): 1D spectra fits file(s)
        gdobj (list): extension name (longslit/multislit) or objID (Echelle)
        order (None or int): order number
        ex_value (str): 'OPT' or 'BOX'
        flux_value (bool): if True it will load fluxed spectra, otherwise load counts

    Returns:
        tuple: Returns the following:
            - waves (ndarray): wavelength array of your spectra, see
              below for the shape information of this array.
            - fluxes (ndarray): flux array of your spectra
            - ivars (ndarray): ivars of your spectra
            - masks (ndarray, bool): mask array of your spectra

        The shapes of all returns are exactly the same.
            - Case 1: np.size(fnames)=np.size(gdobj)=1, order=None for
              Longslit or order=N (an int number) for Echelle
              Longslit/single order for a single fits file, they are 1D
              arrays with the size equal to Nspec
            - Case 2: np.size(fnames)=np.size(gdobj)>1, order=None for
              Longslit or order=N (an int number) for Echelle
              Longslit/single order for a list of fits files, 2D array,
              the shapes are Nspec by Nexp
            - Case 3: np.size(fnames)=np.size(gdobj)=1, order=None All
              Echelle orders for a single fits file, 2D array, the
              shapes are Nspec by Norders
            - Case 4: np.size(fnames)=np.size(gdobj)>1, order=None All
              Echelle orders for a list of fits files, 3D array, the
              shapres are Nspec by Norders by Nexp
    '''

    # read in the first fits file
    if isinstance(fnames, (list, np.ndarray)):
        nexp = np.size(fnames)
        fname0 = fnames[0]
    elif isinstance(fnames, str):
        nexp = 1
        fname0 = fnames

    hdulist = fits.open(fname0)
    header = hdulist[0].header
    npix = header['NPIX']
    pypeline = header['PYPELINE']

    # get the order/slit information
    ntrace0 = np.size(hdulist)-1
    idx_orders = []
    for ii in range(ntrace0):
        idx_orders.append(int(hdulist[ii+1].name.split('-')[1][5:])) # slit ID or order ID


    if pypeline == "Echelle":
        ## np.unique automatically sort the returned array which is not what I want!!!
        ## order_vec = np.unique(idx_orders)
        dum, order_vec_idx = np.unique(idx_orders, return_index=True)
        order_vec = np.array(idx_orders)[np.sort(order_vec_idx)]
        norder = np.size(order_vec)
    else:
        norder = 1

    #TODO This is unneccessarily complicated. The nexp=1 case does the same operations as the nexp > 1 case. Refactor
    # this so that it just does the same set of operations once and then reshapes the array at the end to give you what
    # you want. Let's merge this with unpack orders

    ## Loading data from a single fits file
    if nexp == 1:
        # initialize arrays
        if (order is None) and (pypeline == "Echelle"):
            waves = np.zeros((npix, norder,nexp))
            fluxes = np.zeros_like(waves)
            ivars = np.zeros_like(waves)
            masks = np.zeros_like(waves, dtype=bool)

            for ii, iord in enumerate(order_vec):
                ext_id = gdobj[0]+'-ORDER{:04d}'.format(iord)
                wave_iord, flux_iord, ivar_iord, mask_iord = load_ext_to_array(hdulist, ext_id, ex_value=ex_value,
                                                                               flux_value=flux_value, nmaskedge=nmaskedge)
                waves[:,ii,0] = wave_iord
                fluxes[:,ii,0] = flux_iord
                ivars[:,ii,0] = ivar_iord
                masks[:,ii,0] = mask_iord
        else:
            if pypeline == "Echelle":
                ext_id = gdobj[0]+'-ORDER{:04d}'.format(order)
            else:
                ext_id = gdobj[0]
            waves, fluxes, ivars, masks = load_ext_to_array(hdulist, ext_id, ex_value=ex_value, flux_value=flux_value,
                                                            nmaskedge=nmaskedge)

    ## Loading data from a list of fits files
    else:
        # initialize arrays
        if (order is None) and (pypeline == "Echelle"):
            # store all orders into one single array
            waves = np.zeros((npix, norder, nexp))
        else:
            # store a specific order or longslit
            waves = np.zeros((npix, nexp))
        fluxes = np.zeros_like(waves)
        ivars = np.zeros_like(waves)
        masks = np.zeros_like(waves,dtype=bool)

        for iexp in range(nexp):
            hdulist_iexp = fits.open(fnames[iexp])

            # ToDo: The following part can be removed if all data are reduced using the leatest pipeline
            if pypeline == "Echelle":
                ntrace = np.size(hdulist_iexp) - 1
                idx_orders = []
                for ii in range(ntrace):
                    idx_orders.append(int(hdulist_iexp[ii + 1].name.split('-')[1][5:]))  # slit ID or order ID
                    dum, order_vec_idx = np.unique(idx_orders, return_index=True)
                    order_vec = np.array(idx_orders)[np.sort(order_vec_idx)]
            # ToDo: The above part can be removed if all data are reduced using the leatest pipeline

            if (order is None) and (pypeline == "Echelle"):
                for ii, iord in enumerate(order_vec):
                    ext_id = gdobj[iexp]+'-ORDER{:04d}'.format(iord)
                    wave_iord, flux_iord, ivar_iord, mask_iord = load_ext_to_array(hdulist_iexp, ext_id, ex_value=ex_value,
                                                                                   nmaskedge = nmaskedge, flux_value=flux_value)
                    waves[:,ii,iexp] = wave_iord
                    fluxes[:,ii,iexp] = flux_iord
                    ivars[:,ii,iexp] = ivar_iord
                    masks[:,ii,iexp] = mask_iord
            else:
                if pypeline == "Echelle":
                    ext_id = gdobj[iexp]+'-ORDER{:04d}'.format(order)
                else:
                    ext_id = gdobj[iexp]
                wave, flux, ivar, mask = load_ext_to_array(hdulist_iexp, ext_id, ex_value=ex_value, flux_value=flux_value,
                                                           nmaskedge=nmaskedge)
                waves[:, iexp] = wave
                fluxes[:, iexp] = flux
                ivars[:, iexp] = ivar
                masks[:, iexp] = mask

    return waves, fluxes, ivars, masks, header

def load_spec_order(fname,norder, objid=None, order=None, extract='OPT', flux=True):
    """
    Loading single order spectrum from a PypeIt 1D specctrum fits file.
    it will be called by ech_load_spec

    Args:
        fname (str) : The file name of your spec1d file
        objid (str) : The id of the object you want to load. (default is the first object)
        order (int) : which order you want to load (default is None, loading all orders)
        extract (str) : 'OPT' or 'BOX'
        flux (bool) : default is True, loading fluxed spectra

    Returns:
        XSpectrum1D: spectrum_out
    """
    if objid is None:
        objid = 0
    if order is None:
        msgs.error('Please specify which order you want to load')

    # read extension name into a list
    primary_header = fits.getheader(fname, 0)
    nspec = primary_header['NSPEC']
    extnames = [primary_header['EXT0001']] * nspec
    for kk in range(nspec):
        extnames[kk] = primary_header['EXT' + '{0:04}'.format(kk + 1)]

    # Figure out which extension is the required data
    extnames_array = np.reshape(np.array(extnames),(norder,int(nspec/norder)))
    extnames_good = extnames_array[:,int(objid[3:])-1]
    extname = extnames_good[order]

    try:
        exten = extnames.index(extname) + 1
        msgs.info("Loading extension {:s} of spectrum {:s}".format(extname, fname))
    except:
        msgs.error("Spectrum {:s} does not contain {:s} extension".format(fname, extname))

    spectrum = load_1dspec(fname, exten=exten, extract=extract, flux=flux)
    # Polish a bit -- Deal with NAN, inf, and *very* large values that will exceed
    #   the floating point precision of float32 for var which is sig**2 (i.e. 1e38)
    bad_flux = np.any([np.isnan(spectrum.flux), np.isinf(spectrum.flux),
                       np.abs(spectrum.flux) > 1e30,
                       spectrum.sig ** 2 > 1e10,
                       ], axis=0)
    # Sometimes Echelle spectra have zero wavelength
    bad_wave = spectrum.wavelength < 1000.0*units.AA
    bad_all = bad_flux + bad_wave
    ## trim bad part
    wave_out,flux_out,sig_out = spectrum.wavelength[~bad_all],spectrum.flux[~bad_all],spectrum.sig[~bad_all]
    spectrum_out = XSpectrum1D.from_tuple((wave_out,flux_out,sig_out), verbose=False)
    #if np.sum(bad_flux):
    #    msgs.warn("There are some bad flux values in this spectrum.  Will zero them out and mask them (not ideal)")
    #    spectrum.data['flux'][spectrum.select][bad_flux] = 0.
    #    spectrum.data['sig'][spectrum.select][bad_flux] = 0.

    return spectrum_out

def ech_load_spec(files,objid=None,order=None,extract='OPT',flux=True):
    """
    Loading Echelle spectra from a list of PypeIt 1D spectrum fits files

    Args:
        files (str) : The list of file names of your spec1d file
        objid (str) : The id (one per fits file) of the object you want to load. (default is the first object)
        order (int) : which order you want to load (default is None, loading all orders)
        extract (str) : 'OPT' or 'BOX'
        flux (bool) : default is True, loading fluxed spectra

    Returns:
        XSpectrum1D: spectrum_out
    """

    nfiles = len(files)
    if objid is None:
        objid = ['OBJ0000'] * nfiles
    elif len(objid) == 1:
        objid = objid * nfiles
    elif len(objid) != nfiles:
        msgs.error('The length of objid should be either 1 or equal to the number of spectra files.')

    fname = files[0]
    ext_first = fits.getheader(fname, 1)
    ext_final = fits.getheader(fname, -1)
    norder = abs(ext_final['ECHORDER'] - ext_first['ECHORDER']) + 1
    msgs.info('spectrum {:s} has {:d} orders'.format(fname, norder))
    if norder <= 1:
        msgs.error('The number of orders have to be greater than one for echelle. Longslit data?')

    # Load spectra
    spectra_list = []
    for ii, fname in enumerate(files):

        if order is None:
            msgs.info('Loading all orders into a gaint spectra')
            for iord in range(norder):
                spectrum = load_spec_order(fname, norder, objid=objid[ii],order=iord,extract=extract,flux=flux)
                # Append
                spectra_list.append(spectrum)
        elif order >= norder:
            msgs.error('order number cannot greater than the total number of orders')
        else:
            spectrum = load_spec_order(fname,norder, objid=objid[ii], order=order, extract=extract, flux=flux)
            # Append
            spectra_list.append(spectrum)
    # Join into one XSpectrum1D object
    spectra = collate(spectra_list)
    # Return
    return spectra

def load_sens_dict(filename):
    """
    Load a full (all slit) wv_calib dict

    Includes converting the JSON lists of particular items into ndarray

    Fills self.wv_calib and self.par

    Args:
        filename (str): Master file

    Returns:
        dict or None: self.wv_calib

    """


    # Does the master file exist?
    if not os.path.isfile(filename):
        msgs.warn("No sensfunc file found with filename {:s}".format(filename))
        return None
    else:
        msgs.info("Loading sensfunc from file {:s}".format(filename))
        sens_dict = linetools.utils.loadjson(filename)
        # Recast a few items as arrays
        for key in sens_dict.keys():
            try:
                int(key)
            except ValueError:
                continue
            else:
                for tkey in sens_dict[key].keys():
                    if isinstance(sens_dict[key][tkey], list):
                        sens_dict[key][tkey] = np.array(sens_dict[key][tkey])

        return sens_dict



def waveids(fname):
    infile = fits.open(fname)
    pixels=[]
    msgs.info("Loading fitted arc lines")
    try:
        o = 1
        while True:
            pixels.append(infile[o].data.astype(np.float))
            o+=1
    except:
        pass
    return pixels


def load_multiext_fits(filename, ext):
    """
    Load data and primary header from a multi-extension FITS file

    Args:
        filename (:obj:`str`):
            Name of the file.
        ext (:obj:`str`, :obj:`int`, :obj:`list`):
            One or more file extensions with data to return.  The
            extension can be designated by its 0-indexed integer
            number or its name.

    Returns:
        tuple: Returns the image data from each provided extension.
        If return_header is true, the primary header is also
        returned.
    """
    # Format the input and set the tuple for an empty return
    _ext = ext if isinstance(ext, list) else [ext]
    n_ext = len(_ext)
    # Open the file
    hdu = fits.open(filename)
    head0 = hdu[0].header
    # Only one extension
    if n_ext == 1:
        data = hdu[_ext[0]].data.astype(np.float)
        return data, head0
    # Multiple extensions
    data = tuple([None if hdu[k].data is None else hdu[k].data.astype(np.float) for k in _ext])
    # Return
    return data+(head0,)




"""
Module to run tests on arcoadd
"""
import os

import pytest
import numpy as np

from astropy import units
from linetools.spectra.utils import collate
from linetools.spectra.xspectrum1d import XSpectrum1D

from pypeit.core import coadd
from pypeit.spectrographs.util import load_spectrograph
from pypeit import msgs
from pypeit import utils
from IPython import embed

kast_blue = load_spectrograph('shane_kast_blue')

import warnings
warnings.simplefilter("ignore", UserWarning)

# TODO: Need to rewrite the test for coadd1d. FW commented out most tests at this moment.

def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)

#@pytest.fixture -- Crashes in pytest 4.0.0 as we call this from inside the test, not the call to the test
def dummy_spectrum(s2n=10., rstate=None, seed=1234, wave=None):
    """
    Parameters
    ----------
    s2n
    seed
    wave

    Returns
    -------
    spec : XSpectrum1D

    """
    if rstate is None:
        rstate=np.random.RandomState(seed)
    if wave is None:
        wave = np.linspace(4000., 5000., 2000)
    # Create
    flux = np.ones_like(wave)
    sig = np.ones_like(wave) / s2n
    ispec = XSpectrum1D.from_tuple((wave,flux,sig))
    # Noise and append
    spec = ispec.add_noise(rstate=rstate)
    flux, sig, mask = spec.data['flux'], spec.data['sig'], spec.data['flux'].mask
    ivar = utils.inverse(sig**2)
    return flux, ivar, mask

#@pytest.fixture
def dummy_spectra(s2n=10., seed=1234, wvmnx=None, npix=None):
    """ Generate a set of normalized spectra with varying wavelength
    and noise

    Parameters
    ----------
    s2n : float, optional

    Returns
    -------
    dspec : XSpectrum1D

    """
    rstate=np.random.RandomState(seed)
    if wvmnx is None:
        wvmnx = [[5000., 6000.],
                [4000.5, 5800.5],
                [4500.8, 6300.8],
                ]
    if npix is None:
        npix = [1000, 1001, 1100]
    slist = []
    for ii, inpix in enumerate(npix):
        wave = np.linspace(wvmnx[ii][0], wvmnx[ii][1], inpix)
        #flux = np.ones_like(wave)
        #sig = np.ones_like(wave) / s2n
        #spec = XSpectrum1D.from_tuple((wave,flux,sig))
        ## Noise and append
        #slist.append(spec.add_noise(seed=seed))
        slist.append(dummy_spectrum(wave=wave, s2n=s2n, rstate=rstate))
    # Collate
    dspec = collate(slist, masking='edges')
    #
    return dspec

def test_qa():
    """ Test QA """
    if os.getenv('RUN_ALL_PYPIT_TESTS') is None:
        assert True
        return
    # Setup
    #wvmnx = [[5000., 6000.],
    #         [5000.5, 6000.5],
    #         [5000.8, 6000.8],
    #         ]
    #npix = [1000, 1000, 1000]
    dspec = dummy_spectra(s2n=10.)#, wvmnx=wvmnx, npix=npix)
    dspec.data['flux'][0, 700] *= 1000.  # One bad pixel
    dspec.data['sig'][0, 700] *= 500.
    #TODO rewrite this test
    #coadd.coadd_spectra(dspec, wave_method='concatenate', qafile='tst.pdf')


'''  THIS ARE OLD MODELS
def test_load():
    files = [data_path('spec1d_J2202p1708_KASTb_2015Nov06T024436.08.fits'),
             data_path('spec1d_J2202p1708_KASTb_2015Nov06T031500.09.fits'),
             data_path('spec1d_J2202p1708_KASTb_2015Nov06T034520.54.fits')]
    spectra = coadd.load_spec(files, flux=False)  # Need to add in fluxed test files
    assert spectra.nspec == 3
    # Boxcar too
    spectra = coadd.load_spec(files, extract='box')
'''

'''
Need to rewrite the test for coadd1d.
def test_new_wave_grid():
    # Dummy spectrum
    dspec = dummy_spectra()
    # iref [default]
    iref_wave = coadd1d.new_wave_grid(dspec.data['wave'])
    np.testing.assert_allclose(iref_wave[0], 5000.)
    np.testing.assert_allclose(iref_wave[-1], 6000.)
    # Concatenate
    cat_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='concatenate')
    np.testing.assert_allclose(cat_wave[0], 4000.5)
    np.testing.assert_allclose(cat_wave[-1], 6300.8)
    # Velocity
    vel_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='velocity')
    np.testing.assert_allclose(vel_wave[0], 4000.5)
    np.testing.assert_allclose(vel_wave[-1], 6300.25691664)
    vel_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='velocity', v_pix=100.)
    np.testing.assert_allclose(vel_wave[0], 4000.5)
    np.testing.assert_allclose(vel_wave[-1], 6300.6820934900243)
    # Pixel
    pix_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='pixel', A_pix=2.5)
    np.testing.assert_allclose(pix_wave[0], 4000.5)
    np.testing.assert_allclose(pix_wave[-1], 6303.0)



def test_median_flux():
    """ Test median flux algorithm """
    from pypeit import coadd as arco
    spec = dummy_spectrum(s2n=10)
    # Put in a bad pixel
    spec.data['flux'][0,500] = 0.
    med_flux, std_flux = coadd.median_flux(spec)
    np.testing.assert_allclose(med_flux, 1.0, atol=0.05)  # Noise is random
    np.testing.assert_allclose(std_flux, 0.095, atol=0.004)  # Noise is random
'''

'''
Need to rewrite the test for coadd1d.

def test_median_ratio_flux():
    """ Test median ratio flux algorithm """
    # Setup
    spec1 = dummy_spectrum(s2n=10)
    spec2 = dummy_spectrum(s2n=10)
    spec2.data['flux'][0,:] *= 2.
    spec = collate([spec1,spec2])
    smask = spec.data['sig'].filled(0.) > 0.
    # Put in a bad pixel
    med_flux = coadd1d.median_ratio_flux(spec, smask, 1, 0)
    np.testing.assert_allclose(med_flux, 0.5, atol=0.05)


def test_sn_weight():
    """ Test sn_weight method """
    #  Very low S/N first
    dspec = dummy_spectra(s2n=0.3, seed=1234)
    cat_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='concatenate')
    rspec = dspec.rebin(cat_wave*units.AA, all=True, do_sig=True, masking='none')
    smask = rspec.data['sig'].filled(0.) > 0.
    fluxes, sigs, wave = coadd1d.unpack_spec(rspec)
    ivars = utils.inverse(sigs)
    rms_sn, weights = coadd1d.sn_weights(wave, fluxes, ivars, mask=smask)
    np.testing.assert_allclose(rms_sn[0], 0.318, atol=0.1)  # Noise is random
    #  Low S/N first
    dspec = dummy_spectra(s2n=3., seed=1234)
    cat_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='concatenate')
    rspec = dspec.rebin(cat_wave*units.AA, all=True, do_sig=True, masking='none')
    smask = rspec.data['sig'].filled(0.) > 0.
    fluxes, sigs, wave = coadd1d.unpack_spec(rspec)
    ivars = utils.inverse(sigs)
    rms_sn, weights = coadd1d.sn_weights(wave, fluxes, ivars, mask=smask)
    np.testing.assert_allclose(rms_sn[0], 2.934, atol=0.1)  # Noise is random
    #  High S/N now
    dspec2 = dummy_spectra(s2n=10., seed=1234)
    cat_wave = coadd1d.new_wave_grid(dspec2.data['wave'], wave_method='concatenate')
    rspec2 = dspec2.rebin(cat_wave*units.AA, all=True, do_sig=True, masking='none')
    smask = rspec2.data['sig'].filled(0.) > 0.
    fluxes, sigs, wave = coadd1d.unpack_spec(rspec2)
    ivars = utils.inverse(sigs)
    rms_sn, weights = coadd1d.sn_weights(wave, fluxes, ivars, mask=smask)
    np.testing.assert_allclose(rms_sn[0], 9.904, atol=0.1)  # Noise is random


def test_scale():
    """ Test scale algorithms """
    # Hand
    dspec = dummy_spectra(s2n=10.)
    cat_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='concatenate')
    rspec = dspec.rebin(cat_wave*units.AA, all=True, do_sig=True, masking='none')
    smask = rspec.data['sig'].filled(0.) > 0.
    sv_high = rspec.copy()
    fluxes, sigs, wave = coadd1d.unpack_spec(rspec)
    rms_sn, weights = coadd1d.sn_weights(fluxes, sigs, smask, wave)
    _, _ = coadd1d.scale_spectra(rspec, smask, rms_sn, hand_scale=[3., 5., 10.], scale_method='hand')
    np.testing.assert_allclose(np.median(rspec.flux.value[rspec.sig>0.]), 3., atol=0.01)  # Noise is random
    # Median
    rspec = sv_high.copy()
    fluxes, sigs, wave = coadd1d.unpack_spec(rspec)
    rms_sn, weights = coadd1d.sn_weights(fluxes, sigs, smask, wave)
    _, mthd = coadd1d.scale_spectra(rspec, smask, rms_sn, scale_method='median')
    assert mthd == 'median_flux'
    np.testing.assert_allclose(np.median(rspec.flux.value[rspec.sig>0.]), 1., atol=0.01)  # Noise is random
    #  Auto-none
    dspec = dummy_spectra(s2n=0.1)
    rspec = dspec.rebin(cat_wave*units.AA, all=True, do_sig=True, masking='none')
    fluxes, sigs, wave = coadd1d.unpack_spec(rspec)
    rms_sn, weights = coadd1d.sn_weights(fluxes, sigs, smask, wave)
    an_scls, an_mthd = coadd1d.scale_spectra(rspec, smask, rms_sn)
    assert an_mthd == 'none_SN'
    #  Auto-median
    dspec = dummy_spectra(s2n=1.5)
    rspec = dspec.rebin(cat_wave*units.AA, all=True, do_sig=True, masking='none')
    rspec.data['flux'][1,:] *= 10.
    rspec.data['sig'][1,:] *= 10.
    fluxes, sigs, wave = coadd1d.unpack_spec(rspec)
    rms_sn, weights = coadd1d.sn_weights(fluxes, sigs, smask, wave)
    am_scls, am_mthd = coadd1d.scale_spectra(rspec, smask, rms_sn, scale_method='median')
    assert am_mthd == 'median_flux'
    np.testing.assert_allclose(am_scls[1], 0.1, atol=0.01)


def test_grow_mask():
    """ Test grow_mask method.  Now works on 1d spectra"""
    # Setup
    dspec = dummy_spectrum(s2n=10.)
    mask = np.ma.getmaskarray(dspec.data['wave'][0,:])
    mask[:] = True
    # Set some
    mask[0] = False
    mask[50] = False
    mask[500] = False
    # Grow
    new_mask = coadd1d.grow_mask(mask, n_grow=1)
    # Test
    badp = np.where(new_mask == False)[0]
    assert np.all(badp == np.array([0,1,49,50,51,499,500,501]))
    #badpb = np.where(new_mask[1,:])[0]
    #assert np.all(badpb == np.array([0,1]))
    #badpc = np.where(new_mask[2,:])[0]
    #assert np.all(badpc == np.array([1098,1099]))
    # Grow 2
    new_mask2 = coadd1d.grow_mask(mask, n_grow=2)
    badp2 = np.where(new_mask2 == False)[0]
    assert len(badp2) == 13


def test_1dcoadd():
    """ Test 1dcoadd method"""
    # Setup
    dspec = dummy_spectra(s2n=10.)
    cat_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='concatenate')
    rspec = dspec.rebin(cat_wave*units.AA, all=True, do_sig=True, masking='none')
    smask = rspec.data['sig'].filled(0.) > 0.
    fluxes, sigs, wave = coadd1d.unpack_spec(rspec)
    rms_sn, weights = coadd1d.sn_weights(fluxes, sigs, smask, wave)
    # Coadd
    spec1d = coadd1d.one_d_coadd(rspec, smask, weights)
    assert spec1d.npix == 1740

def test_cleancr():
    """ Test clean CR method"""
    # Setup
    dspec = dummy_spectra(s2n=10.)
    dspec.data['flux'][0, 700] *= 1000.  # One bad pixel
    dspec.data['sig'][0, 700] *= 500.
    cat_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='concatenate')
    rspec = dspec.rebin(cat_wave*units.AA, all=True, do_sig=True, masking='none')
    #
    smask = rspec.data['sig'].filled(0.) <= 0.
    coadd1d.clean_cr(rspec, smask)


def test_coadd():
    """ Test full coadd method"""
    # Setup
    dspec = dummy_spectra(s2n=10.)
    dspec.data['flux'][0, 700] *= 1000.  # One bad pixel
    dspec.data['sig'][0, 700] *= 500.
    spec1d = coadd1d.coadd_spectra(kast_blue, None, dspec, wave_grid_method='concatenate')
    assert np.isclose(np.median(spec1d.flux.value), 1., atol=0.003)


# TODO: This needs to be fixed.
def test_coadd_with_fluxing():
    """ Test full coadd method with flux scaling"""
    scale_dict = dict(filter='DECAM-R', mag=19.0, mag_type='AB')
    # Setup
    dspec = dummy_spectra(s2n=10.)
    dspec.data['flux'][0, 700] *= 1000.  # One bad pixel
    dspec.data['sig'][0, 700] *= 500.
    spec1d = coadd1d.coadd_spectra(kast_blue, None, dspec, wave_grid_method='concatenate', flux_scale=scale_dict)
    # Test
    assert np.median(spec1d.flux.value) > 6.61


def test_coadd_qa():
    if os.getenv('PYPIT') is None:
        assert True
        return
'''

'''
def test_sigma_clip():
    """ Test sigma_clip method """
    from pypeit import coadd as arco
    # Setup
    dspec = dummy_spectra(s2n=10.)
    cat_wave = coadd1d.new_wave_grid(dspec.data['wave'], wave_method='concatenate')
    # Rebin
    rspec = dspec.rebin(cat_wave*units.AA, all=True, do_sig=True)
    sn2, weights = coadd1d.sn_weight(cat_wave, rspec.data['flux'], rspec.data['sig']**2)
    # Here we go
    rspec.data['flux'][0, 700] = 999.
    final_mask = coadd1d.sigma_clip(rspec.data['flux'], rspec.data['sig']**2, sn2=sn2)


'''




from configobj import ConfigObj
import numpy as np
from pypeit import par, msgs
from astropy.io import fits
from pypeit.par import pypeitpar
from pypeit.spectrographs.util import load_spectrograph
import argparse
from IPython import embed
import textwrap
from pypeit import sensfunc
import os


# A trick from stackoverflow to allow multi-line output in the help:
#https://stackoverflow.com/questions/3853722/python-argparse-how-to-insert-newline-in-the-help-text
class SmartFormatter(argparse.HelpFormatter):

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)

def read_sensfile(ifile):
    """
    Read a PypeIt sens file, akin to a standard PypeIt file

    The top is a config block that sets ParSet parameters
      The spectrograph is not required

    Args:
        ifile: str
          Name of the flux file

    Returns:
        spectrograph: Spectrograph
        cfg_lines: list
          Config lines to modify ParSet values
        flux_dict: dict
          Contains spec1d_files and flux_files
          Empty if no flux block is specified

    """

    # Read in the pypeit reduction file
    msgs.info('Loading the fluxcalib file')
    lines = par.util._read_pypeit_file_lines(ifile)

    # Construct config to get spectrograph
    cfg_lines = list(lines)
    #cfg = ConfigObj(cfg_lines)
    #spectrograph_name = cfg['rdx']['spectrograph']
    #spectrograph = load_spectrograph(spectrograph_name)

    # Return
    return cfg_lines



# TODO Need an option here for multi_spec_det detectors, passed as a list of numbers in the SensFunc parset, or as --det 3 7 on the command line
def parser(options=None):
    parser = argparse.ArgumentParser(description='Parse', formatter_class=SmartFormatter)
    parser.add_argument("spec1dfile", type=str,
                        help="spec1d file for the standard that will be used to compute sensitivity function")
    parser.add_argument("--algorithm", type=str, default=None, choices=['UVIS', 'IR'],
                        help="R|Override the default algorithm for computing the sensitivity function. \n"
                             "Note that it is not possible to set --algorithm and simultaneously use a .sens file with\n"
                             "the --sens_file option. If you are using a .sens file set the algorithm there via:\n"
                             "\n"
                             "    [sensfunc]\n"
                             "         algorithm = IR\n"
                             "\n"
                             "The algorithm options are:\n"
                             "\n"
                             "    UVIS = Should be used for data with lambda < 7000A.\n" 
                             "    No detailed model of telluric absorption but corrects for atmospheric extinction.\n"
                             "\n"
                             "    IR   = Should be used for data with lambbda > 7000A.\n"
                             "    Peforms joint fit for sensitivity function and telluric absorption using HITRAN models.\n"
                             "\n")
    parser.add_argument("--multi", type=str,
                        help="R|List of detector numbers to splice together for instruments with multiple detectors\n"
                             "arranged in the spectral direction, e.g. --multi = '3,7'\n"
                             "Note that it is not possible to set --multi and \n"
                             "simultaneously use a .sens file with the --sens_file option.\n"
                             "If you are using a .sens file set the multi_spec_det param there via:\n"
                             "\n"
                             "         [sensfunc]\n"
                             "              multi_spec_det = 3,7\n"
                             "\n")
    parser.add_argument("-o", "--outfile", type=str,
                        help="Ouput file for sensitivity function. If not specified, the sensitivity function will "
                             "be written out to a standard filename in the current working directory, i.e. if the "
                             "standard spec1d file is named spec1d_b24-Feige66_KASTb_foo.fits the sensfunc will be "
                             "written to sens_b24-Feige66_KASTb_foo.fits")
    parser.add_argument("-s", "--sens_file", type=str, help="Configuration file to change default sensivity function parameters")
    parser.add_argument("--debug", default=False, action="store_true", help="show debug plots?")
    parser.add_argument("--debug_init", default=False, action="store_true",
                        help="debug the initilization of the sensfunc + telluric fit for the IR algorithm")
    #parser.add_argument("--plot", default=False, action="store_true", help="Show the sensitivity function?")
    parser.add_argument("--par_outfile", default='sensfunc.par', help="Name of outut file to save the parameters used by the fit")

    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)
    return args



def main(args):
    """ Executes sensitivity function computation.
    """

    # Check parameter inputs
    if args.algorithm is not None and args.sens_file is not None:
        msgs.error("It is not possible to set --algorithm and simultaneously use a .sens file via\n"
                   "the --sens_file option. If you are using a .sens file set the algorithm there via:\n"
                   "\n"
                   "    [sensfunc]\n"
                   "         algorithm = IR\n"
                   "\n")

    if args.multi is not None and args.sens_file is not None:
        msgs.error("It is not possible to set --multi and simultaneously use a .sens file via\n"
                   "the --sens_file option. If you are using a .sens file set the detectors there via:\n"
                   "\n"
                   "\n"
                   "         [sensfunc]\n"
                   "              multi_spec_det = 3,7\n"
                   "\n")
    # Determine the spectrograph
    header = fits.getheader(args.spec1dfile)
    spectrograph = load_spectrograph(header['PYP_SPEC'])
    spectrograph_def_par = spectrograph.default_pypeit_par()
    # If the .sens file was passed in read it and overwrite default parameters
    if args.sens_file is not None:
        cfg_lines = read_sensfile(args.sens_file)
        par = pypeitpar.PypeItPar.from_cfg_lines(cfg_lines=spectrograph_def_par.to_config(),
                                                 merge_with=cfg_lines)
    else:
        par = spectrograph_def_par

    # If algorithm was provided override defaults. Note this does undo .sens file since they cannot both be passed
    if args.algorithm is not None:
        par['sensfunc']['algorithm'] = args.algorithm
    # If multi was set override defaults. Note this does undo .sens file since they cannot both be passed
    if args.multi is not None:
        # parse
        multi_spec_det  = [int(item) for item in args.multi.split(',')]
        par['sensfunc']['multi_spec_det'] = multi_spec_det

    # TODO Add parsing of detectors here. If detectors passed from the command line, overwrite the parset values read
    # in from the .sens file

    # Write the par to disk
    print("Writing the parameters to {}".format(args.par_outfile))
    par['sensfunc'].to_config('sensfunc.par', section_name='sensfunc', include_descr=False)
    # TODO JFH I would like to be able to run only par['sensfunc'].to_config('sensfunc.par') but this crashes.

    # Parse the output filename
    outfile = (os.path.basename(args.spec1dfile)).replace('spec1d','sens') if args.outfile is None else args.outfile
    # Instantiate the relevant class for the requested algorithm
    sensobj = sensfunc.SensFunc.get_instance(args.spec1dfile, outfile, par=par['sensfunc'], debug=args.debug)
    # Generate the sensfunc
    sensobj.run()
    # Write it out to a file
    sensobj.save()

    #TODO JFH Add a show_sensfunc option here and to the sensfunc classes.


#!/usr/bin/env python
#
# See top-level LICENSE file for Copyright information
#
# -*- coding: utf-8 -*-
"""
This script generates files to setup a PypeIt run
"""
import argparse
from pypeit.spectrographs.util import valid_spectrographs

def parser(options=None):
    # TODO: Add argument that specifies the log file
    parser = argparse.ArgumentParser(description="Script to setup a PypeIt run [v3]")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-r', '--root', type=str, default=None,
                       help='File path+root, e.g. /data/Kast/b ')
    #group.add_argument('-p', '--pypeit_file', type=str, default=None,
    #                   help='PypeIt file to use')
    parser.add_argument('-s', '--spectrograph', default=None, type=str,
                        help='A valid spectrograph identifier: {0}'.format(
                                ', '.join(valid_spectrographs())))
    parser.add_argument('-e', '--extension', default='.fits',
                        help='File extension; compression indicators (e.g. .gz) not required.')
    parser.add_argument('-d', '--output_path', default=None,
                        help='Path to top-level output directory.  '
                             'Default is the current working directory.')
    parser.add_argument('-o', '--overwrite', default=False, action='store_true',
                        help='Overwrite any existing files/directories')
    parser.add_argument('-c', '--cfg_split', default=None, type=str,
                        help='Generate the PypeIt files and folders by input configuration. [all or A,B or B,D,E or E]')
    parser.add_argument('-b', '--background', default=False, action='store_true',
                        help='Include the background-pair columns for the user to edit')
    parser.add_argument('-v', '--verbosity', type=int, default=2,
                        help='Level of verbosity from 0 to 2; default is 2.')
#    parser.add_argument('-q', '--quick', default=False, help='Quick reduction',
#                        action='store_true')
#    parser.add_argument('-c', '--cpus', default=False,
#                        help='Number of CPUs for parallel processing', action='store_true')
#    parser.print_help()

    return parser.parse_args() if options is None else parser.parse_args(options)


def main(args):

    import os
    import pdb as debugger

    from pypeit import msgs
    from pypeit.pypeitsetup import PypeItSetup

    # Check that the spectrograph is provided if using a file root
    if args.root is not None:
        if args.spectrograph is None:
            raise ValueError('Must provide spectrograph identifier with file root.')
        # Check that input spectrograph is supported
        instruments_served = valid_spectrographs()
        if args.spectrograph not in instruments_served:
            raise ValueError('Instrument \'{0}\' unknown to PypeIt.\n'.format(args.spectrograph)
                             + '\tOptions are: {0}\n'.format(', '.join(instruments_served))
                             + '\tSelect an available instrument or consult the documentation '
                             + 'on how to add a new instrument.')

    # Get the output directory
    output_path = os.getcwd() if args.output_path is None else args.output_path
    sort_dir = os.path.join(output_path, 'setup_files')
   
    # Initialize PypeItSetup based on the arguments
    if args.root is not None:
        ps = PypeItSetup.from_file_root(args.root, args.spectrograph, extension=args.extension,
                                        output_path=sort_dir)
    else:
        # Should never reach here
        raise IOError('Need to set -r !!')

    # Run the setup
    ps.run(setup_only=True, sort_dir=sort_dir, write_bkg_pairs=args.background)

    # Use PypeItMetaData to write the complete PypeIt file
    if args.cfg_split is not None:
        pypeit_file = os.path.join(output_path, '{0}.pypeit'.format(args.spectrograph))
        config_list = [item.strip() for item in args.cfg_split.split(',')]
        ps.fitstbl.write_pypeit(pypeit_file, cfg_lines=ps.user_cfg, write_bkg_pairs=args.background,
                                configs=config_list)
    return 0




"""
This script allows the user to add/delete/modify object traces

.. todo::

    Implement color scaling with RMB click+drag

"""

import os, sys
import copy
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.widgets import Button, RadioButtons
from scipy.interpolate import RectBivariateSpline

from pypeit import specobjs
from pypeit import msgs

operations = dict({'cursor': "Select object trace (LMB click)\n" +
                   "         Navigate (LMB drag = pan, RMB drag = zoom)\n" +
                   "         Note: In order to pan/zoom, you need to first activate\n" +
                   "         the pan/zoom tool with the 'p' key, or by selecting the\n" +
                   "         pan/zoom tool on the Matplotlib navigation tool menu. You\n" +
                   "         can also zoom using the magnifying glass (select this option\n" +
                   "         from the Matplotlib navigation tool menu)\n",
                   'a': "Add a new object trace using the selected method",
                   'c': "Center the window at the location of the mouse",
                   'd': "Delete selected object trace",
                   'h/r': "Return zoom to the original plotting limits",
                   'p': "Toggle pan/zoom with the cursor",
                   '?': "Display the available options",
                   })
#                   Commands that could be used for manual object location
#                   'c': "Clear the anchor points and start again",
#                   'm': "Insert a fitting anchor for manual object tracing",
#                   'n': "Delete the fitting anchor nearest to the cursor",
#                   '+/-': "Raise/Lower the order of the fitting polynomial"


class ObjectTraces:
    """
    Simple class to store the object traces
    """

    def __init__(self):
        self._det = []
        self._add_rm = []  # Flag to say if an object was been added (1), removed (-1), or was there originally (0)
        self._pos_spat = []
        self._pos_spec = []
        self._trace_spat = []
        self._trace_spec = []
        self._fwhm = []

    @property
    def nobj(self):
        return len(self._det)

    def from_specobj(self, sobjs, det):
        """Fill the object traces from a SpecObjs class

        Args:
            sobjs (SpecObjs): An instance of the SpecObjs class
            det (int): Detector to which this applies
        """
        nobj = sobjs.nobj
        for ii in range(nobj):
            pos_spat = sobjs[ii].trace_spat[sobjs[ii].trace_spat.size//2]
            pos_spec = sobjs[ii].trace_spec[sobjs[ii].trace_spec.size//2]
            self.add_object(det, pos_spat, pos_spec, sobjs[ii].trace_spat, sobjs[ii].trace_spec, sobjs[ii].fwhm, addrm=0)

    def from_dict(self, obj_dict, det):
        """Fill the object traces from a SpecObjs class

        Args:
            obj_dict (dict): A dictionary containing the spatial object traces and FWHM
            det (int): Detector to which this applies
        """
        nobj = len(obj_dict['traces'])
        for ii in range(nobj):
            spec_trace = np.arange(obj_dict['traces'][ii].size)
            pos_spat = obj_dict['traces'][ii][obj_dict['traces'][ii].size//2]
            pos_spec = spec_trace[spec_trace.size//2]
            self.add_object(det, pos_spat, pos_spec, obj_dict['traces'][ii], spec_trace, obj_dict['fwhm'][ii], addrm=0)

    def add_object(self, det, pos_spat, pos_spec, trc_spat, trc_spec, fwhm, addrm=1):
        """Add an object trace

        Args:
            det (int): Detector to add a slit on
            pos_spat (float): Spatial pixel position
            pos_spec (float): Spectral pixel position
            trc_spat (ndarray): Spatial trace of object
            trc_spec (ndarray): Spectral trace of object
            fwhm (float): FWHM of the object
            addrm (int): Flag to say if an object was been added (1), removed (-1), or was an auto found slit (0)
        """
        self._det.append(det)
        self._add_rm.append(addrm)
        self._pos_spat.append(pos_spat)
        self._pos_spec.append(pos_spec)
        self._trace_spat.append(trc_spat)
        self._trace_spec.append(trc_spec)
        self._fwhm.append(fwhm)

    def delete_object(self, ind):
        """Delete an object trace

        Args:
            ind (int): Index of object trace to remove
        """
        if self._add_rm[ind] == 1:
            # If an object was added by hand, the user can delete it from the list
            del self._det[ind]
            del self._add_rm[ind]
            del self._pos_spat[ind]
            del self._pos_spec[ind]
            del self._trace_spat[ind]
            del self._trace_spec[ind]
            del self._fwhm[ind]
        else:
            # otherwise, just flag that it needs to be removed
            self._add_rm[ind] = -1

    def get_pypeit_string(self):
        """Construct a string that can be placed in the .pypeit file to define new object traces
        """
        strall = ""
        for ii in range(self.nobj):
            if self._add_rm[ii] == 1:
                strall += ",{0:d}:{1:.1f}:{2:.1f}:{3:.1f}".format(self._det[ii], self._pos_spat[ii], self._pos_spec[ii], self._fwhm[ii])
        strall += "\n"
        return strall[1:]


class ObjFindGUI(object):
    """
    GUI to interactively identify object traces. The GUI can be run within
    PypeIt during data reduction, or as a standalone script outside of
    PypeIt. To initialise the GUI, call the initialise() function in this
    file.
    """

    def __init__(self, canvas, image, frame, det, sobjs, trace_dict, axes, profdict,
                 slit_ids=None, printout=False, runtime=False):
        """Controls for the interactive Object ID tasks in PypeIt.

        The main goal of this routine is to interactively add/delete/modify
        object apertures.

        Args:
            canvas (Matploltib figure canvas): The canvas on which all axes are contained
            image (AxesImage): The image plotted to screen
            frame (ndarray): The image data
            det (int): Detector to add a slit on
            sobjs (SpecObjs, None): An instance of the SpecObjs class
            trace_dict (dict): A dictionary containing information about the object traces
            axes (dict): Dictionary of four Matplotlib axes instances (Main spectrum panel, two for residuals, one for information)
            profdict (dict): Dictionary containing profile information (profile data, and the left/right lines displayinf the FWHM)
            slit_ids (list, None): List of slit ID numbers
            printout (bool): Should the results be printed to screen
            runtime (bool): Is the GUI being launched during data reduction?
        """
        # Store the axes
        self._det = det
        self.image = image
        self.frame = frame
        self.nspec, self.nspat = frame.shape[0], frame.shape[1]
        self._spectrace = np.arange(self.nspec)
        self.profile = profdict
        self._printout = printout
        self._runtime = runtime
        self._slit_ids = slit_ids
        self._trcdict = trace_dict
        self.axes = axes
        self.specobjs = sobjs
        self.objtraces = []
        self._object_traces = ObjectTraces()
        self.anchors = []
        self._obj_idx = -1
        self._spatpos = np.arange(frame.shape[1])[np.newaxis, :].repeat(frame.shape[0], axis=0)  # Spatial coordinate (as the frame shape)
        self.empty_mantrace()
        if sobjs is None:
            self._object_traces.from_dict(self._trcdict['objtrc'], det)
        else:
            self._object_traces.from_specobj(sobjs, det)

        # Unset some of the matplotlib keymaps
        matplotlib.pyplot.rcParams['keymap.fullscreen'] = ''        # toggling fullscreen (Default: f, ctrl+f)
        #matplotlib.pyplot.rcParams['keymap.home'] = ''              # home or reset mnemonic (Default: h, r, home)
        matplotlib.pyplot.rcParams['keymap.back'] = ''              # forward / backward keys to enable (Default: left, c, backspace)
        matplotlib.pyplot.rcParams['keymap.forward'] = ''           # left handed quick navigation (Default: right, v)
        #matplotlib.pyplot.rcParams['keymap.pan'] = ''              # pan mnemonic (Default: p)
        matplotlib.pyplot.rcParams['keymap.zoom'] = ''              # zoom mnemonic (Default: o)
        matplotlib.pyplot.rcParams['keymap.save'] = ''              # saving current figure (Default: s)
        matplotlib.pyplot.rcParams['keymap.quit'] = ''              # close the current figure (Default: ctrl+w, cmd+w)
        matplotlib.pyplot.rcParams['keymap.grid'] = ''              # switching on/off a grid in current axes (Default: g)
        matplotlib.pyplot.rcParams['keymap.yscale'] = ''            # toggle scaling of y-axes ('log'/'linear') (Default: l)
        matplotlib.pyplot.rcParams['keymap.xscale'] = ''            # toggle scaling of x-axes ('log'/'linear') (Default: L, k)
        matplotlib.pyplot.rcParams['keymap.all_axes'] = ''          # enable all axes (Default: a)

        # Initialise the main canvas tools
        canvas.mpl_connect('draw_event', self.draw_callback)
        canvas.mpl_connect('button_press_event', self.button_press_callback)
        canvas.mpl_connect('key_press_event', self.key_press_callback)
        canvas.mpl_connect('button_release_event', self.button_release_callback)
        canvas.mpl_connect('motion_notify_event',  self.mouse_move_callback)
        self.canvas = canvas

        # Interaction variables
        self._respreq = [False, None]  # Does the user need to provide a response before any other operation will be permitted? Once the user responds, the second element of this array provides the action to be performed.
        self._qconf = False  # Confirm quit message
        self._changes = False
        self._use_updates = True
        self._trcmthd = 'object'
        self.mmx, self.mmy = 0, 0
        self._inslit = 0  # Which slit is the mouse in

        # Draw the spectrum
        self.canvas.draw()

        # Initialise buttons and menu options
        self._ax_meth_default = 'Object'
        self._methdict = dict({'Object': [0, 'object'],
                               'Standard Star': [1, 'std'],
                               'Slit Edges': [2, 'slit']})
#                               'Manual': [3, 'manual']
        self.initialise_menu()

    def print_help(self):
        """Print the keys and descriptions that can be used for Identification
        """
        keys = operations.keys()
        print("===============================================================")
        print("Add/remove object traces until you are happy with the resulting")
        print("traces. When you've finished, click one of the exit buttons on")
        print("the right side of the page. If you click 'Continue (and save changes)'")
        print("the object traces will be printed to the terminal, where you can")
        print("copy them into your .pypeit file.")
        print("")
        print("thick coloured dashed lines = object traces")
        print("thick coloured solid line   = currently selected object trace")
        print("thin green/blue lines       = slit edges")
        print("")
        print("Meanings of the different coloured dashed lines:")
        print(" green = user-defined object trace")
        print(" blue  = trace automatically generated with PypeIt")
        print(" red   = trace automatically generated with PypeIt (deleted)")
        print("===============================================================")
        print("       OBJECT ID OPERATIONS")
        for key in keys:
            print("{0:6s} : {1:s}".format(key, operations[key]))
        print("---------------------------------------------------------------")

    def initialise_menu(self):
        """Initialise the menu buttons
        """
        axcolor = 'lightgoldenrodyellow'
        # Continue with reduction (using updated specobjs)
        ax_cont = plt.axes([0.82, 0.85, 0.15, 0.05])
        self._ax_cont = Button(ax_cont, "Continue (and save changes)", color=axcolor, hovercolor='y')
        self._ax_cont.on_clicked(self.button_cont)
        # Continue with reduction (using original specobjs)
        ax_exit = plt.axes([0.82, 0.79, 0.15, 0.05])
        self._ax_exit = Button(ax_exit, "Continue (don't save changes)", color=axcolor, hovercolor='y')
        self._ax_exit.on_clicked(self.button_exit)
        # Button to select trace method
        rax = plt.axes([0.82, 0.59, 0.15, 0.15], facecolor=axcolor)
        rax.set_title("Select trace method:")
        self._ax_meth = RadioButtons(rax, ('Object', 'Standard Star', 'Slit Edges'))#, 'Manual'))
        self._ax_meth.on_clicked(self.radio_meth)
        # Determine the best default to use:
        if self._trcdict["trace_model"]["object"]["trace_model"] is not None:
            self._ax_meth_default = 'Object'
        elif self._trcdict["trace_model"]["std"]["trace_model"] is not None:
            self._ax_meth_default = 'Standard Star'
        elif self._trcdict["trace_model"]["slit"]["trace_model"] is not None:
            self._ax_meth_default = 'Slit Edges'
#        elif self._trcdict["trace_model"]["manual"]["trace_model"] is not None:
#            self._ax_meth_default = 'Manual'
        # Set the active method
        self._ax_meth.set_active(self._methdict[self._ax_meth_default][0])

    def radio_meth(self, label, infobox=True):
        """Tell the code what to do when a different trace method is selected

        Args:
            label (str): The label of the radio button that was clicked
        """
        # Update the radio button
        if self._methdict[label][1]:
            self._trcmthd = self._methdict[label][1]
        # Check if the method is available, if not, change to the default (manual is always allowed)
        if self._trcmthd != "manual":
            if self._trcdict["trace_model"][self._trcmthd]["trace_model"] is None:
                self.update_infobox(message="That option is not available - changing to default", yesno=False)
                self._ax_meth.set_active(self._methdict[self._ax_meth_default][0])
                self._trcmthd = self._methdict[self._ax_meth_default][1]
            else:
                if infobox:
                    self.update_infobox(message="Trace method set to: {0:s}".format(label), yesno=False)
        else:
            if infobox:
                self.update_infobox(message="Trace method set to: {0:s}".format(label), yesno=False)

    def button_cont(self, event):
        """What to do when the 'exit and save' button is clicked
        """
        self._respreq = [True, "exit_update"]
        self.update_infobox(message="Are you sure you want to exit and use the updated object traces?", yesno=True)

    def button_exit(self, event):
        """What to do when the 'exit and do not save changes' button is clicked
        """
        self._respreq = [True, "exit_restore"]
        self.update_infobox(message="Are you sure you want to exit and use the original object traces?", yesno=True)

    def replot(self):
        """Redraw the entire canvas
        """
        self.canvas.restore_region(self.background)
        self.draw_objtraces()
        self.draw_anchors()
        self.canvas.draw()

    def draw_objtraces(self):
        """Draw the object traces
        """
        for i in self.objtraces: i.pop(0).remove()
        self.objtraces = []
        # Plot the object traces
        allcols = ['DodgerBlue', 'LimeGreen', 'r']  # colors mean: [original, added, deleted]
        for iobj in range(self._object_traces.nobj):
            color = allcols[self._object_traces._add_rm[iobj]]
            if iobj == self._obj_idx:
                self.objtraces.append(self.axes['main'].plot(self._object_traces._trace_spat[iobj],
                                                             self._object_traces._trace_spec[iobj],
                                                             color=color,
                                                             linestyle='-', linewidth=4, alpha=0.5))
            else:
                self.objtraces.append(self.axes['main'].plot(self._object_traces._trace_spat[iobj],
                                                             self._object_traces._trace_spec[iobj],
                                                             color=color,
                                                             linestyle='--', linewidth=3, alpha=0.5))

    def draw_anchors(self):
        """Draw the anchors for manual tracing
        """
        for i in self.anchors: i.pop(0).remove()
        self.anchors = []
        # Plot the best fitting trace, if it exists
        if self._mantrace["spat_trc"] is not None:
            self.anchors.append(self.axes['main'].plot(self._mantrace["spat_trc"], self._mantrace["spec_trc"],
                                                       'g-', linewidth=3, alpha=0.5))
        # Plot the anchor points on top
        self.anchors.append(self.axes['main'].plot(self._mantrace["spat_a"], self._mantrace["spec_a"], 'ro', alpha=0.5))

    def draw_profile(self):
        """Draw the object profile
        """
        if self._obj_idx == -1:
            sz = self.profile['profile'].get_xdata.size
            self.profile['profile'].set_ydata(np.zeros(sz))
        else:
            # Plot the extent of the FWHM
            self.profile['fwhm'][0].set_xdata(-self._object_traces._fwhm[self._obj_idx]/2.0)
            self.profile['fwhm'][1].set_xdata(+self._object_traces._fwhm[self._obj_idx]/2.0)
            # Update the data shown
            objprof = self.make_objprofile()
            self.profile['profile'].set_ydata(objprof)
            self.axes['profile'].set_xlim([-self._object_traces._fwhm[self._obj_idx],
                                           +self._object_traces._fwhm[self._obj_idx]])
            omin, omax = objprof.min(), objprof.max()
            self.axes['profile'].set_ylim([omin-0.1*(omax-omin), omax+0.1*(omax-omin)])

    def draw_callback(self, event):
        """Draw callback (i.e. everytime the canvas is being drawn/updated)

        Args:
            event (Event): A matplotlib event instance
        """
        # Get the background
        self.background = self.canvas.copy_from_bbox(self.axes['main'].bbox)
        self.draw_objtraces()

    def get_ind_under_point(self, event):
        """Get the index of the object trace closest to the cursor

        Args:
            event (Event): Matplotlib event instance containing information about the event
        """
        mindist = self._spatpos.shape[0]**2
        self._obj_idx = -1
        for iobj in range(self._object_traces.nobj):
            dist = (event.xdata-self._object_traces._trace_spat[iobj])**2 +\
                   (event.ydata-self._object_traces._trace_spec[iobj])**2
            if np.min(dist) < mindist:
                mindist = np.min(dist)
                self._obj_idx = iobj
        if self._obj_idx != -1:
            self.draw_profile()
        return

    def get_axisID(self, event):
        """Get the ID of the axis where an event has occurred

        Args:
            event (Event): Matplotlib event instance containing information about the event

        Returns:
            int, None: Axis where the event has occurred
        """
        if event.inaxes == self.axes['main']:
            return 0
        elif event.inaxes == self.axes['info']:
            return 1
        elif event.inaxes == self.axes['profile']:
            return 2
        return None

    def mouse_move_callback(self, event):
        """Store the locations of mouse as it moves across the canvas
        """
        if event.inaxes is None:
            return
        axisID = self.get_axisID(event)
        if event.inaxes == self.axes['main']:
            self.mmx, self.mmy = event.xdata, event.ydata

    def button_press_callback(self, event):
        """What to do when the mouse button is pressed

        Args:
            event (Event): Matplotlib event instance containing information about the event
        """
        if event.inaxes is None:
            return
        if self.canvas.toolbar.mode != "":
            return
        if event.button == 1:
            self._addsub = 1
        elif event.button == 3:
            self._addsub = 0
        if event.inaxes == self.axes['main']:
            self._start = [event.x, event.y]
        elif event.inaxes == self.axes['profile']:
            self._start = [event.x, event.y]

    def button_release_callback(self, event):
        """What to do when the mouse button is released

        Args:
            event (Event): Matplotlib event instance containing information about the event
        """
        if event.inaxes is None:
            return
        if event.inaxes == self.axes['info']:
            if (event.xdata > 0.8) and (event.xdata < 0.9):
                answer = "y"
            elif event.xdata >= 0.9:
                answer = "n"
            else:
                return
            self.operations(answer, -1)
            self.update_infobox(default=True)
            return
        elif event.inaxes == self.axes['profile']:
            if (event.x == self._start[0]) and (event.y == self._start[1]):
                self.set_fwhm(event.xdata)
                self.update_infobox(message="FWHM updated for the selected object", yesno=False)
            return
        elif self._respreq[0]:
            # The user is trying to do something before they have responded to a question
            return
        if self.canvas.toolbar.mode != "":
            return
        # Draw an actor
        axisID = self.get_axisID(event)
        if axisID is not None:
            if axisID <= 2:
                self._end = [event.x, event.y]
                if (self._end[0] == self._start[0]) and (self._end[1] == self._start[1]):
                    # The mouse button was pressed (not dragged)
                    self.get_ind_under_point(event)
                    self.update_infobox(message="Object selected", yesno=False)
                    pass
                elif self._end != self._start:
                    # The mouse button was dragged
                    if axisID == 0:
                        if self._start > self._end:
                            tmp = self._start
                            self._start = self._end
                            self._end = tmp
                        # Now do something
                        pass
        # Now plot
        self.canvas.restore_region(self.background)
        self.draw_objtraces()
        self.canvas.draw()

    def key_press_callback(self, event):
        """What to do when a key is pressed

        Args:
            event (Event): Matplotlib event instance containing information about the event
        """
        # Check that the event is in an axis...
        if not event.inaxes:
            return
        # ... but not the information box!
        if event.inaxes == self.axes['info']:
            return
        axisID = self.get_axisID(event)
        self.operations(event.key, axisID)

    def operations(self, key, axisID):
        """Canvas operations

        Args:
            key (str): Which key has been pressed
            axisID (int): The index of the axis where the key has been pressed (see get_axisID)
        """
        # Check if the user really wants to quit
        if key == 'q' and self._qconf:
            if self._changes:
                self.update_infobox(message="WARNING: There are unsaved changes!!\nPress q again to exit", yesno=False)
                self._qconf = True
            else:
                msgs.bug("Need to change this to kill and return the results to PypeIt")
                plt.close()
        elif self._qconf:
            self.update_infobox(default=True)
            self._qconf = False

        # Manage responses from questions posed to the user.
        if self._respreq[0]:
            if key != "y" and key != "n":
                return
            else:
                # Switch off the required response
                self._respreq[0] = False
                # Deal with the response
                if self._respreq[1] == "delete_object" and key == "y":
                    self.delete_object()
                elif self._respreq[1] == "clear_anchors" and key == "y":
                    self.empty_mantrace()
                    self.replot()
                elif self._respreq[1] == "exit_update" and key == "y":
                    self._use_updates = True
                    self.print_pypeit_info()
                    self.operations("qu", None)
                elif self._respreq[1] == "exit_restore" and key == "y":
                    self._use_updates = False
                    self.operations("qr", None)
                else:
                    return
            # Reset the info box
            self.update_infobox(default=True)
            return

        if key == '?':
            self.print_help()
        elif key == 'a':
            self.add_object()
        elif key == 'c':
            if axisID == 0:
                # If this is pressed on the main window
                self.recenter()
#        elif key == 'c':
#            self._respreq = [True, "clear_anchors"]
#            self.update_infobox(message="Are you sure you want to clear the anchors", yesno=True)
        elif key == 'd':
            if self._obj_idx != -1:
                self._respreq = [True, "delete_object"]
                self.update_infobox(message="Are you sure you want to delete this object trace", yesno=True)
#        elif key == 'm':
#            if self._trcmthd != 'manual':
#                self.update_infobox(message="To add an anchor point, set the 'manual' trace method", yesno=False)
#            else:
#                self.add_anchor()
#        elif key == 'n':
#            self.remove_anchor()
        elif key == 'qu' or key == 'qr':
            if self._changes:
                self.update_infobox(message="WARNING: There are unsaved changes!!\nPress q again to exit", yesno=False)
                self._qconf = True
            else:
                plt.close()
#        elif key == '+':
#            if self._mantrace["polyorder"] < 10:
#                self._mantrace["polyorder"] += 1
#                self.update_infobox(message="Polynomial order = {0:d}".format(self._mantrace["polyorder"]), yesno=False)
#                self.fit_anchors()
#            else:
#                self.update_infobox(message="Polynomial order must be <= 10", yesno=False)
#        elif key == '-':
#            if self._mantrace["polyorder"] > 1:
#                self._mantrace["polyorder"] -= 1
#                self.update_infobox(message="Polynomial order = {0:d}".format(self._mantrace["polyorder"]), yesno=False)
#                self.fit_anchors()
#            else:
#                self.update_infobox(message="Polynomial order must be >= 1", yesno=False)
        self.replot()

    def add_anchor(self):
        """Add a manual anchor point
        """
        self._mantrace['spat_a'].append(self.mmx)
        self._mantrace['spec_a'].append(self.mmy)
        self.fit_anchors()

    def remove_anchor(self):
        """Remove a manual anchor point
        """
        # Find the anchor closest to the mouse position
        if len(self._mantrace['spat_a']) != 0:
            mindist = (self._mantrace['spat_a'][0]-self.mmx)**2 + (self._mantrace['spec_a'][0]-self.mmy)**2
            minidx = 0
            for ianc in range(1, len(self._mantrace['spat_a'])):
                dist = (self._mantrace['spat_a'][ianc] - self.mmx) ** 2 + (self._mantrace['spec_a'][ianc] - self.mmy) ** 2
                if dist < mindist:
                    mindist = dist
                    minidx = ianc
            del self._mantrace['spat_a'][minidx]
            del self._mantrace['spec_a'][minidx]
        self.fit_anchors()

    def fit_anchors(self):
        """Fit the manual anchor points
        """
        if len(self._mantrace['spat_a']) <= self._mantrace['polyorder']:
            self.update_infobox(message="You need to select more trace points before manually adding\n" +
                                        "a manual object trace. To do this, use the 'm' key", yesno=False)
        else:
            # Fit a polynomial to the anchor points
            coeff = np.polyfit(self._mantrace['spec_a'], self._mantrace['spat_a'], self._mantrace['polyorder'])
            self._mantrace['spat_trc'] = np.polyval(coeff, self._mantrace['spec_trc'])
        # Replot, regardless of whether a fit is done (a point might have been added/removed)
        self.replot()

    def get_slit(self):
        """Find the slit that the mouse is currently in
        """
        ypos = int(self.mmy)
        for sl in range(self._trcdict['slit_left'].shape[1]):
            if (self.mmx > self._trcdict['slit_left'][ypos, sl]) and (self.mmx < self._trcdict['slit_righ'][ypos, sl]):
                self._inslit = sl
                return

    def add_object(self):
        if self._trcmthd == 'manual' and len(self._mantrace['spat_a']) <= self._mantrace['polyorder']:
            self.update_infobox(message="You need to select more trace points before manually adding\n" +
                                        "a manual object trace. To do this, use the 'm' key", yesno=False)
            return
        # Add an object trace
        spec_vec = self._mantrace['spec_trc']
        if self._trcmthd == 'manual':
            trace_model = self._mantrace['spat_trc'].copy()
            # Now empty the manual tracing
            self.empty_mantrace()
        else:
            if self._trcmthd == 'slit':
                self.get_slit()
                trace_model = self._trcdict["trace_model"][self._trcmthd]["trace_model"][:, self._inslit].copy()
            else:
                trace_model = self._trcdict["trace_model"][self._trcmthd]["trace_model"].copy()
            spat_0 = np.interp(self.mmy, spec_vec, trace_model)
            shift = self.mmx - spat_0
            trace_model += shift
        # Determine the FWHM
        if self._object_traces.nobj != 0:
            fwhm = self._object_traces._fwhm[0]
        else:  # Otherwise just use the fwhm parameter input to the code (or the default value)
            fwhm = 2

        # Finally, add this object to the list
        self._object_traces.add_object(self._det, self.mmx, self.mmy, trace_model, self._spectrace.copy(), fwhm)

    def add_object_sobj(self):
        """Add an object to specobjs
        """
        if self._trcmthd == 'manual' and len(self._mantrace['spat_a']) <= self._mantrace['polyorder']:
            self.update_infobox(message="You need to select more trace points before manually adding\n" +
                                        "a manual object trace. To do this, use the 'm' key", yesno=False)
            return
        # Add an object trace
        spec_vec = self._mantrace['spec_trc']
        if self._trcmthd == 'manual':
            trace_model = self._mantrace['spat_trc'].copy()
            # Now empty the manual tracing
            self.empty_mantrace()
        else:
            trace_model = self._trcdict["trace_model"][self._trcmthd]["trace_model"].copy()
            spat_0 = np.interp(self.mmy, spec_vec, trace_model)
            shift = self.mmx - spat_0
            trace_model += shift
        xsize = self._trcdict["slit_righ"] - self._trcdict["slit_left"]
        nsamp = np.ceil(xsize.max())
        # Extract the SpecObj parameters
        par = self._trcdict['sobj_par']
        # Create a SpecObj
        thisobj = specobjs.SpecObj(par['frameshape'], par['slit_spat_pos'], par['slit_spec_pos'],
                                   det=par['det'], setup=par['setup'], slitid=par['slitid'],
                                   orderindx=par['orderindx'], objtype=par['objtype'])
        thisobj.hand_extract_spat = self.mmx
        thisobj.hand_extract_spec = self.mmy
        thisobj.hand_extract_det = par['det']
        thisobj.hand_extract_fwhm = None
        thisobj.hand_extract_flag = True
        f_ximg = RectBivariateSpline(spec_vec, np.arange(self.nspat), par["ximg"])
        thisobj.spat_fracpos = f_ximg(thisobj.hand_extract_spec, thisobj.hand_extract_spat,
                                      grid=False)  # interpolate from ximg
        thisobj.smash_peakflux = np.interp(thisobj.spat_fracpos * nsamp, np.arange(nsamp),
                                           self._trcdict['profile'])  # interpolate from fluxconv
        # assign the trace
        thisobj.trace_spat = trace_model
        thisobj.trace_spec = spec_vec
        thisobj.spat_pixpos = thisobj.trace_spat[self.nspec//2]
        thisobj.set_idx()
        if self._object_traces.nobj != 0:
            thisobj.fwhm = self._object_traces._fwhm[0]
        else:  # Otherwise just use the fwhm parameter input to the code (or the default value)
            thisobj.fwhm = 2
        # Finally, add new object
        self.specobjs.add_sobj(thisobj)

    def delete_object(self):
        """Delete an object trace
        """
        self._object_traces.delete_object(self._obj_idx)
        self._obj_idx = -1
        self.replot()

    def delete_object_sobj(self):
        """Delete a specobj
        """
        self.specobjs.remove_sobj(self._obj_idx)
        self._obj_idx = -1

    def print_pypeit_info(self):
        """print text that the user should insert into their .pypeit file
        """
        if 1 in self._object_traces._add_rm:
            msgs.info("Include the following info in the manual_extract column in your .pypeit file:\n")
            print(self._object_traces.get_pypeit_string())

    def recenter(self):
        xlim = self.axes['main'].get_xlim()
        ylim = self.axes['main'].get_ylim()
        xmin = self.mmx - 0.5*(xlim[1]-xlim[0])
        xmax = self.mmx + 0.5*(xlim[1]-xlim[0])
        ymin = self.mmy - 0.5*(ylim[1]-ylim[0])
        ymax = self.mmy + 0.5*(ylim[1]-ylim[0])
        self.axes['main'].set_xlim([xmin, xmax])
        self.axes['main'].set_ylim([ymin, ymax])

    def make_objprofile(self):
        """Generate an object profile from the traces
        """
        coords = self._spatpos - self._object_traces._trace_spat[self._obj_idx][:, np.newaxis]
        ww = np.where(np.abs(coords) < 4*self._object_traces._fwhm[self._obj_idx])
        bincent = self.profile['profile'].get_xdata()
        offs = 0.5*(bincent[1]-bincent[0])
        edges = np.append(bincent[0]-offs, bincent+offs)
        prof, _ = np.histogram(coords[ww], bins=edges, weights=self.frame[ww])
        return prof/ww[0].size

    def set_fwhm(self, xdata):
        """Set the FWHM using the available panel

        Args:
            xdata (float): The x coordinate selected by the user
        """
        self._object_traces._fwhm[self._obj_idx] = 2.0*np.abs(xdata)
        self.draw_profile()
        self.replot()
        return

    def get_specobjs(self):
        """Get the updated version of SpecObjs

        Returns:
            SpecObjs: SpecObjs Class
        """
        if self._use_updates:
            msgs.work("Have not updated SpecObjs yet")
            return self.specobjs
        else:
            return None

    def update_infobox(self, message="Press '?' to list the available options",
                       yesno=True, default=False):
        """Send a new message to the information window at the top of the canvas

        Args:
            message (str): Message to be displayed
        """
        self.axes['info'].clear()
        if default:
            self.axes['info'].text(0.5, 0.5, "Press '?' to list the available options", transform=self.axes['info'].transAxes,
                          horizontalalignment='center', verticalalignment='center')
            self.canvas.draw()
            return
        # Display the message
        self.axes['info'].text(0.5, 0.5, message, transform=self.axes['info'].transAxes,
                      horizontalalignment='center', verticalalignment='center')
        if yesno:
            self.axes['info'].fill_between([0.8, 0.9], 0, 1, facecolor='green', alpha=0.5, transform=self.axes['info'].transAxes)
            self.axes['info'].fill_between([0.9, 1.0], 0, 1, facecolor='red', alpha=0.5, transform=self.axes['info'].transAxes)
            self.axes['info'].text(0.85, 0.5, "YES", transform=self.axes['info'].transAxes,
                          horizontalalignment='center', verticalalignment='center')
            self.axes['info'].text(0.95, 0.5, "NO", transform=self.axes['info'].transAxes,
                          horizontalalignment='center', verticalalignment='center')
        self.axes['info'].set_xlim((0, 1))
        self.axes['info'].set_ylim((0, 1))
        self.canvas.draw()

    def empty_mantrace(self):
        """Generate an empty dictionary for the manual tracing
        """
        self._mantrace = dict(spat_a=[], spec_a=[], spat_trc=None, spec_trc=np.arange(self.nspec), polyorder=0)
        return


def initialise(det, frame, trace_dict, sobjs, slit_ids=None, runtime=False, printout=False):
    """Initialise the 'ObjFindGUI' window for interactive object tracing

        Args:
            frame (ndarray): Sky subtracted science image
            trace_dict (dict, None): Dictionary containing slit and object trace information
            sobjs (SpecObjs, None): SpecObjs Class
            det (int): Detector index
            slit_ids (list, None): List of slit ID numbers
            printout (bool): Should the results be printed to screen
            runtime (bool): Is this GUI being launched during a data reduction?

        Returns:
            ObjFindGUI: Returns an instance of the ObjFindGUI class
    """

    # This allows the input lord and rord to either be (nspec, nslit) arrays or a single
    # vectors of size (nspec)
    if trace_dict['slit_left'].ndim == 2:
        nslit = trace_dict['slit_left'].shape[1]
    else:
        nslit = 1
        trace_dict['slit_left'] = trace_dict['slit_left'].reshape((trace_dict['slit_left'].size, 1))
        trace_dict['slit_righ'] = trace_dict['slit_righ'].reshape((trace_dict['slit_righ'].size, 1))
    lordloc = trace_dict['slit_left']
    rordloc = trace_dict['slit_righ']

    if trace_dict["trace_model"]["slit"]["trace_model"].ndim == 1:
        arr = trace_dict["trace_model"]["slit"]["trace_model"]
        trace_dict["trace_model"]["slit"]["trace_model"] = arr.reshape((arr.size, 1))


    # Assign the slit IDs if none were provided
    if slit_ids is None:
        slit_ids = [str(slit) for slit in np.arange(nslit)]

    # Determine the scale of the image
    med = np.median(frame)
    mad = np.median(np.abs(frame-med))
    vmin = med-3*mad
    vmax = med+3*mad

    # Add the main figure axis
    fig, ax = plt.subplots(figsize=(16, 9), facecolor="white")
    plt.subplots_adjust(bottom=0.05, top=0.85, left=0.05, right=0.8)
    image = ax.imshow(frame, aspect='auto', cmap = 'Greys', vmin=vmin, vmax=vmax)

    # Overplot the slit traces
    specarr = np.arange(lordloc.shape[0])
    for sl in range(nslit):
        ax.plot(lordloc[:, sl], specarr, 'g-')
        ax.plot(rordloc[:, sl], specarr, 'b-')

    # Add an object profile axis
    axprof = fig.add_axes([0.82, 0.05, .15, 0.30])
    profx = np.arange(-10, 10.1, 0.1)
    profy = np.zeros(profx.size)
    profile = axprof.plot(profx, profy, 'k-')
    vlinel = axprof.axvline(-1, color='r')
    vliner = axprof.axvline(+1, color='r')
    axprof.set_title("Object profile")
    axprof.set_xlim((-3, 3))
    axprof.set_ylim((0, 1))
    axprof.set_yticks([])

    # Add an information GUI axis
    axinfo = fig.add_axes([0.15, .92, .7, 0.07])
    axinfo.get_xaxis().set_visible(False)
    axinfo.get_yaxis().set_visible(False)
    axinfo.text(0.5, 0.5, "Press '?' to list the available options", transform=axinfo.transAxes,
                horizontalalignment='center', verticalalignment='center')
    axinfo.set_xlim((0, 1))
    axinfo.set_ylim((0, 1))

    axes = dict(main=ax, profile=axprof, info=axinfo)
    profdict = dict(profile=profile[0], fwhm=[vlinel, vliner])
    # Initialise the object finding window and display to screen
    fig.canvas.set_window_title('PypeIt - Object Tracing')
    ofgui = ObjFindGUI(fig.canvas, image, frame, det, sobjs, trace_dict, axes, profdict,
                       slit_ids=slit_ids, printout=printout, runtime=runtime)
    plt.show()

    return ofgui



"""
Module to run tests on WaveCalib class
Requires files in Development suite and an Environmental variable
"""
import os

import pytest
import glob
import numpy as np

from astropy.table import Table

from pypeit import wavecalib
from pypeit.core.wavecal import waveio
from pypeit.tests.tstutils import dev_suite_required, cooked_required
from pypeit.spectrographs import util

@cooked_required
def test_user_redo():
    # Check for files
    spectrograph = util.load_spectrograph('shane_kast_blue')

    # Instantiate
    par = spectrograph.default_pypeit_par()['calibrations']['wavelengths']
    master_dir = os.path.join(os.getenv('PYPEIT_DEV'), 'Cooked', 'WaveCalib')
    master_key = 'ShaneKastBlue_A'
    waveCalib = wavecalib.WaveCalib(None, None, spectrograph, par, master_dir=master_dir,
                                    master_key=master_key, reuse_masters=True)
    assert os.path.isfile(waveCalib.master_file_path), 'Did not find Cooked file.'

    wv_calib = waveCalib.load()
    # Setup
    waveCalib.par['sigdetect'] = 5.
    nslit = 1
    waveCalib.make_maskslits(nslit)
    npix = len(waveCalib.wv_calib['0']['spec'])
    waveCalib.arccen = np.zeros((npix,nslit))
    waveCalib.arccen[:,0] = waveCalib.wv_calib['0']['spec']
    # Do it
    new_wv_calib = waveCalib.build_wv_calib(waveCalib.arccen, 'holy-grail', skip_QA=True)
    # Test
    assert new_wv_calib['0']['rms'] < 0.2
    # Now also test the utility script that reads in the wavecalib
    #wv_calib_load = wavecalib.WaveCalib.load_from_file(waveCalib.master_file_path)
    wv_calib_load = waveio.load_wavelength_calibration(waveCalib.master_file_path)
    assert np.all(wv_calib['0']['fitc'] == wv_calib_load['0']['fitc'])


# TODO: Bring back some of these tests...

'''
def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files/wavecalib')
    return os.path.join(data_dir, filename)


@pytest.fixture
@cooked_required
def master_dir():
    return os.path.join(os.getenv('PYPEIT_DEV'), 'Cooked', 'MF_shane_kast_blue')


def read_old_fitstbl(spectrograph, f):
    fitstbl = PypeItMetaData(spectrograph, data=Table.read(f))
    type_bits = np.zeros(len(fitstbl), dtype=fitstbl.bitmask.minimum_dtype())
    for bit in fitstbl.bitmask.keys():
        type_bits[fitstbl[bit]] = fitstbl.bitmask.turn_on(type_bits[fitstbl[bit]], flag=bit)
    fitstbl.set_frame_types(type_bits)
    return fitstbl


@dev_suite_required
def test_step_by_step(master_dir):
    root_path = os.path.join(os.getenv('PYPEIT_DEV'), 'Cooked', 'MF')
    setup = 'A_01_aa'

    # Load up the Masters
    AImg = arcimage.ArcImage('shane_kast_blue', setup=setup, master_dir=master_dir, mode='reuse')
    msarc, header, _ = AImg.load_master_frame()
    TSlits = traceslits.TraceSlits.from_master_files(os.path.join(master_dir,
                                                                  'MasterTrace_A_01_aa'))
    TSlits._make_pixel_arrays()

    # Kludge to read in old fitstbl object into new one
    fitstbl = read_old_fitstbl(AImg.spectrograph,
                               os.path.join(master_dir, 'shane_kast_blue_setup_A.fits'))

    # Instantiate, note I'm not setting the bad pixel mask to keep this simple here
    waveCalib = wavecalib.WaveCalib(msarc, spectrograph='shane_kast_blue', setup=setup,
                                    master_dir=master_dir, mode='reuse', fitstbl=fitstbl,
                                    sci_ID=1, det=1)

    # Extract arcs
    arccen, maskslits = waveCalib._extract_arcs(TSlits.lcen, TSlits.rcen, TSlits.slitpix)
    assert arccen.shape == (2048, 1)
    # Arcparam
    arcparam = waveCalib._load_arcparam()
    assert isinstance(arcparam, dict)
    # wv_calib
    wv_calib = waveCalib._build_wv_calib('arclines', skip_QA=True)
    assert isinstance(wv_calib, dict)
    # Master
    # This doesn't save steps nor arcparam which *is* done in the master() call
    waveCalib.save_master(wv_calib, outfile=data_path('tmp.json'))


@dev_suite_required
def test_one_shot(master_dir):
    root_path = os.path.join(os.getenv('PYPEIT_DEV'), 'Cooked', 'MF')
    setup = 'A_01_aa'

    # Load up the Masters
    AImg = arcimage.ArcImage('shane_kast_blue', setup=setup, master_dir=master_dir, mode='reuse')
    msarc, header, _ = AImg.load_master_frame()
    TSlits = traceslits.TraceSlits.from_master_files(os.path.join(master_dir,
                                                                  'MasterTrace_A_01_aa'))
    TSlits._make_pixel_arrays()

    # Kludge to read in old fitstbl object into new one
    fitstbl = read_old_fitstbl(AImg.spectrograph,
                               os.path.join(master_dir, 'shane_kast_blue_setup_A.fits'))

    # Do it
    waveCalib = wavecalib.WaveCalib(msarc, spectrograph='shane_kast_blue', setup=setup,
                                    master_dir=master_dir, fitstbl=fitstbl, sci_ID=1, det=1)
    wv_calib2, _ = waveCalib.run(TSlits.lcen, TSlits.rcen, TSlits.slitpix, skip_QA=True)
    #
    assert 'arcparam' in wv_calib2.keys()


# JFH I think this test should be discarded. Ultimately if the dev suite runs it is a much more powerful test of the functionality
# of arclines. Furthermore, if things jiggle around a bit in any of the algorithms, then this test breaks but that is not necessarily
# a failure. It just means the nubmber of lines or quality of the solution changed a bit.
def test_wavecalib_general():

    # LRISb 600/4000 with the longslit
    names = ['LRISb_600_4000_longslit']
    spec_files = ['lrisb_600_4000_PYPIT.json']
    all_lines = [['CdI', 'HgI', 'ZnI']]
    all_wvcen = [4400.]
    all_disp = [1.26]
    fidxs = [0]
    scores = [dict(rms=0.1, nxfit=13, nmatch=10)]

    # LRISb 400/3400 with the longslit
    names += ['LRISb_400_3400_longslit']
    spec_files += ['lrisb_400_3400_PYPIT.json']
    all_lines += [['NeI', 'ArI', 'CdI', 'KrI', 'XeI', 'ZnI', 'HgI']]
    all_wvcen += [4400.]
    all_disp += [1.26]
    fidxs += [0]
    scores += [dict(rms=0.1, nxfit=13, nmatch=10)]


    # LRISb off-center
    # NeI lines are seen beyond the dicrhoic but are ignored
#    names += ['LRISb_600_4000_red']
#    src_files += ['LRISb_600_LRX.hdf5']  # Create with low_redux.py if needed
#    all_wvcen += [5000.]
#    all_disp += [1.26]
#    all_lines += [['CdI', 'ZnI', 'HgI']]  #,'NeI']]
#    fidxs += [18]
#    scores += [dict(rms=0.08, nxfit=10, nmatch=10)]


    # LRISr 600/7500 longslit
    names += ['LRISr_600_7500_longslit']
    spec_files += ['lrisr_600_7500_PYPIT.json']
    all_wvcen += [7000.]
    all_disp += [1.6]
    all_lines += [['ArI', 'HgI', 'KrI', 'NeI', 'XeI']]
    fidxs += [-1]
    scores += [dict(rms=0.08, nxfit=30, nmatch=50)]

    # LRISr 900/XX00 longslit -- blue
#    names += ['LRISr_900_XX00_longslit']
#    src_files += ['lrisr_900_XX00_PYPIT.json']
#    all_wvcen += [5800.]
#    all_disp += [1.08]
#    all_lines += [['ArI', 'HgI', 'KrI', 'NeI', 'XeI', 'CdI', 'ZnI']]
#    fidxs += [-1]
#    scores += [dict(rms=0.08, nxfit=10, nmatch=10)]


    # LRISr 400/8500 longslit -- red
    names += ['LRISr_400_8500_longslit']
    spec_files += ['lrisr_400_8500_PYPIT.json']
    all_wvcen += [8000.]
    all_disp += [2.382]
    all_lines += [['ArI', 'HgI', 'KrI', 'NeI', 'XeI']]
    fidxs += [-1]
    scores += [dict(rms=0.1, nxfit=39, nmatch=40)]

    # Kastb 600 grism
    names += ['KASTb_600_standard']
    spec_files += ['kastb_600_PYPIT.json']
    all_lines += [['CdI', 'HeI', 'HgI']]
    all_wvcen += [4400.]
    all_disp += [1.02]
    fidxs += [0]
    scores += [dict(rms=0.1, nxfit=13, nmatch=8)]

    # Kastr 600/7500 grating
    names += ['KASTr_600_7500_standard']
    spec_files += ['kastr_600_7500_PYPIT.json']
    all_lines += [['ArI', 'NeI', 'HgI']]
    all_wvcen += [6800.]
    all_disp += [2.345]
    fidxs += [0]
    scores += [dict(rms=0.1, nxfit=10, nmatch=8)]

    # Keck DEIMOS
    names += ['keck_deimos_830g_l']
    spec_files += ['keck_deimos_830g_l_PYPIT.json']
    all_lines += [['ArI', 'NeI', 'KrI', 'XeI']]
    all_wvcen += [7450.]
    all_disp += [0.467]
    fidxs += [0]
    scores += [dict(rms=0.1, nxfit=20, nmatch=20)]


    # Favored parameters (should match those in the defaults)
    min_nsig = 20.
    nonlinear_counts = 5.6e4 # It woudl be better if we actually dug this out of the spectrograph class for each instrument

    # Run it
    for name, spec_file, lines, wvcen, disp, score, fidx in zip(
            names, spec_files, all_lines, all_wvcen, all_disp, scores, fidxs):
        # Load spectrum
        exten = spec_file.split('.')[-1]
        if exten == 'json':
            with open(data_path(spec_file), 'r') as f:
                pypit_fit = json.load(f)
            try:
                # Old format
                spec = np.array(pypit_fit['spec'])
            except KeyError:
                spec = np.array(pypit_fit['0']['spec'])
        elif exten == 'hdf5':
            hdf = h5py.File(data_path(spec_file), 'r')
            spec = hdf['arcs/{:d}/spec'.format(fidx)].value

        arcfitter = autoid.General(spec.reshape((spec.size, 1)), lines, min_nsig=min_nsig,
                                   rms_threshold=score['rms'], nonlinear_counts=nonlinear_counts)
        final_fit = arcfitter._all_final_fit

        # Score
        grade = True
        slit = '0'
        if final_fit[slit]['rms'] > score['rms']:
            grade = False
            print("Solution for {:s} failed RMS!!".format(name))
        if len(final_fit[slit]['xfit']) < score['nxfit']:
            grade = False
            print("Solution for {:s} failed N xfit!!".format(name))
#        if patt_dict[slit]['nmatch'] < score['nmatch']:
#            grade = False
#            print("Solution for {:s} failed N match!!".format(name))
        assert grade

'''



"""
Module to run tests on WaveTilts class
Requires files in Development suite and an Environmental variable
"""
import os

import pytest
import numpy as np


from pypeit.tests.tstutils import dev_suite_required, load_kast_blue_masters, cooked_required
from pypeit import wavetilts
from pypeit.core import tracewave, pixels
from pypeit.par import pypeitpar
from pypeit.spectrographs.util import load_spectrograph


def data_path(filename):
    data_dir = os.path.join(os.path.dirname(__file__), 'files')
    return os.path.join(data_dir, filename)


@pytest.fixture
@cooked_required
def master_dir():
    return os.path.join(os.getenv('PYPEIT_DEV'), 'Cooked', 'Shane_Kast_blue')

@cooked_required
def test_instantiate_from_master(master_dir):
    master_file = os.path.join(os.getenv('PYPEIT_DEV'), 'Cooked', 'Shane_Kast_blue',
                               'MasterTilts_A_1_01.fits')
    waveTilts = wavetilts.WaveTilts.from_master_file(master_file)
    assert isinstance(waveTilts.tilts_dict, dict)

@cooked_required
def test_step_by_step(master_dir):
    # Masters
    spectrograph = load_spectrograph('shane_kast_blue')
    msarc, edges = load_kast_blue_masters(aimg=True, edges=True)
    # Instantiate
    master_key = 'A_1_01'
    parset = spectrograph.default_pypeit_par()
    par = parset['calibrations']['tilts']
    wavepar = parset['calibrations']['wavelengths']
    waveTilts = wavetilts.WaveTilts(msarc, edges.convert_to_tslits_dict(), spectrograph, par,
                                    wavepar, det=1, master_key=master_key, master_dir=master_dir,
                                    reuse_masters=True)
    # Extract arcs
    arccen, arccen_bpm, maskslits = waveTilts.extract_arcs()
    #waveTilts.slitcen, waveTilts.slitmask, waveTilts.inmask)
    assert arccen.shape == (2048,1)
    # Tilts in the slit
    slit = 0
    waveTilts.slitmask = pixels.tslits2mask(waveTilts.tslits_dict)
    thismask = waveTilts.slitmask == slit
    waveTilts.lines_spec, waveTilts.lines_spat \
            = waveTilts.find_lines(arccen[:, slit], waveTilts.slitcen[:, slit], slit)

    trcdict = waveTilts.trace_tilts(waveTilts.msarc.image, waveTilts.lines_spec,
                                    waveTilts.lines_spat, thismask, slit)
    assert isinstance(trcdict, dict)
    # 2D Fit
    spat_order = waveTilts._parse_param(waveTilts.par, 'spat_order', slit)
    spec_order = waveTilts._parse_param(waveTilts.par, 'spec_order', slit)
    coeffs = waveTilts.fit_tilts(trcdict, thismask, waveTilts.slitcen[:, slit], spat_order,
                                 spec_order,slit, doqa=False)
    tilts = tracewave.fit2tilts(waveTilts.slitmask_science.shape, coeffs, waveTilts.par['func2d'])
    assert np.max(tilts) < 1.01


@cooked_required
def test_run(master_dir):
    # Masters
    spectrograph = load_spectrograph('shane_kast_blue')
    msarc, edges = load_kast_blue_masters(aimg=True, edges=True)
    # Instantiate
    master_key = 'A_1_01'
    spectrograph.detector[0]['saturation'] = 60000.
    spectrograph.detector[0]['nonlinear'] = 0.9
    par = pypeitpar.WaveTiltsPar()
    wavepar = pypeitpar.WavelengthSolutionPar()
    waveTilts = wavetilts.WaveTilts(msarc, edges.convert_to_tslits_dict(), spectrograph, par,
                                    wavepar, det=1, master_key=master_key, master_dir=master_dir,
                                    reuse_masters=True)
    # Run
    tilts_dict, mask = waveTilts.run(doqa=False)
    assert isinstance(tilts_dict['tilts'], np.ndarray)




""" Class to generate an image from one or more files (and other pieces).
"""

import inspect

import os
import numpy as np


from pypeit import msgs

from pypeit.core import combine
from pypeit.par import pypeitpar
from pypeit import utils

from pypeit.images import pypeitimage
from pypeit.images import processrawimage
from pypeit.images import rawimage
from pypeit.images import maskimage

from IPython import embed


class CombineImage(object):
    """
    Class to generate an image from one or more files (and other pieces).

    The core processing steps are handled by ProcessRawImage
    This object is mainly for combining multiple images

    Args:
        spectrograph (:class:`pypeit.spectrographs.spectrograph.Spectrograph`):
            Spectrograph used to take the data.
        det (:obj:`int`, optional):
            The 1-indexed detector number to process.
        par (:class:`pypeit.par.pypeitpar.ProcessImagesPar`):
            Parameters that dictate the processing of the images.  See
            :class:`pypeit.par.pypeitpar.ProcessImagesPar` for the
            defaults.

    """
    def __init__(self, spectrograph, det, par, files):

        # Required parameters
        self.spectrograph = spectrograph
        self.det = det
        if not isinstance(par, pypeitpar.ProcessImagesPar):
            msgs.error('Provided ParSet for must be type ProcessImagesPar.')
        self.par = par  # This musts be named this way as it is frequently a child
        self.files = files
        if self.nfiles == 0:
            msgs.error('Combineimage requires a list of files to instantiate')

    def process_one(self, filename, process_steps, bias, pixel_flat=None, illum_flat=None, bpm=None):
        """
        Process a single image

        Args:
            filename (str):
                File to process
            process_steps (list):
                List of processing steps
            bias (np.ndarray or None):
                Bias image
            pixel_flat (np.ndarray, optional):
                Flat image
            illum_flat (np.ndarray, optional):
                Illumination image
            bpm (np.ndarray, optional):
                Bad pixel mask

        Returns:
            :class:`pypeit.images.pypeitimage.PypeItImage`:

        """
        # Load raw image
        rawImage = rawimage.RawImage(filename, self.spectrograph, self.det)
        # Process
        processrawImage = processrawimage.ProcessRawImage(rawImage, self.par, bpm=bpm)
        processedImage = processrawImage.process(process_steps, bias=bias, pixel_flat=pixel_flat,
                                                 illum_flat=illum_flat)
        # Return
        return processedImage

    def run(self, process_steps, bias, pixel_flat=None, illum_flat=None,
            ignore_saturation=False, sigma_clip=True, bpm=None, sigrej=None, maxiters=5):
        """
        Generate a PypeItImage from a list of images

        Mainly a wrapper to coadd2d.weighted_combine()

        This may also generate the ivar, crmask, rn2img and mask

        Args:
            process_steps (list):
            bias (np.ndarray or None):
                Bias image or instruction
            pixel_flat (np.ndarray, optional):
                Flat image
            illum_flat (np.ndarray, optional):
                Illumination image
            sigma_clip (bool, optional):
                Perform sigma clipping
            sigrej (int or float, optional): Rejection threshold for sigma clipping.
                 Code defaults to determining this automatically based on the number of images provided.
            maxiters (int, optional):
                Number of iterations for the clipping
            bpm (np.ndarray, optional):
                Bad pixel mask.  Held in ImageMask
            ignore_saturation (bool, optional):
                If True, turn off the saturation flag in the individual images before stacking
                This avoids having such values set to 0 which for certain images (e.g. flat calibrations)
                can have unintended consequences.

        Returns:
            :class:`pypeit.images.pypeitimage.PypeItImage`:

        """
        # Loop on the files
        nimages = len(self.files)
        for kk, ifile in enumerate(self.files):
            # Process a single image
            pypeitImage = self.process_one(ifile, process_steps, bias, pixel_flat=pixel_flat,
                                           illum_flat=illum_flat, bpm=bpm)
            # Are we all done?
            if len(self.files) == 1:
                return pypeitImage
            elif kk == 0:
                # Get ready
                shape = (nimages, pypeitImage.bpm.shape[0], pypeitImage.bpm.shape[1])
                img_stack = np.zeros(shape)
                ivar_stack= np.zeros(shape)
                rn2img_stack = np.zeros(shape)
                crmask_stack = np.zeros(shape, dtype=bool)
                # Mask
                bitmask = maskimage.ImageBitMask()
                mask_stack = np.zeros(shape, bitmask.minimum_dtype(asuint=True))
            # Process
            img_stack[kk,:,:] = pypeitImage.image
            # Construct raw variance image and turn into inverse variance
            if pypeitImage.ivar is not None:
                ivar_stack[kk, :, :] = pypeitImage.ivar
            else:
                ivar_stack[kk, :, :] = 1.
            # Mask cosmic rays
            if pypeitImage.crmask is not None:
                crmask_stack[kk, :, :] = pypeitImage.crmask
            # Read noise squared image
            if pypeitImage.rn2img is not None:
                rn2img_stack[kk, :, :] = pypeitImage.rn2img
            # Final mask for this image
            # TODO This seems kludgy to me. Why not just pass ignore_saturation to process_one and ignore the saturation
            # when the mask is actually built, rather than untoggling the bit here
            if ignore_saturation:  # Important for calibrations as we don't want replacement by 0
                indx = pypeitImage.bitmask.flagged(pypeitImage.mask, flag=['SATURATION'])
                pypeitImage.mask[indx] = pypeitImage.bitmask.turn_off(pypeitImage.mask[indx], 'SATURATION')
            mask_stack[kk, :, :] = pypeitImage.mask

        # Coadd them
        weights = np.ones(nimages)/float(nimages)
        img_list = [img_stack]
        var_stack = utils.inverse(ivar_stack)
        var_list = [var_stack, rn2img_stack]
        img_list_out, var_list_out, outmask, nused = combine.weighted_combine(
            weights, img_list, var_list, (mask_stack == 0),
            sigma_clip=sigma_clip, sigma_clip_stack=img_stack, sigrej=sigrej, maxiters=maxiters)

        # Build the last one
        final_pypeitImage = pypeitimage.PypeItImage(img_list_out[0],
                                                    ivar=utils.inverse(var_list_out[0]),
                                                    bpm=pypeitImage.bpm,
                                                    rn2img=var_list_out[1],
                                                    crmask=np.invert(outmask),
                                                    binning=pypeitImage.binning)
        nonlinear_counts = self.spectrograph.nonlinear_counts(self.det,
                                                              apply_gain='apply_gain' in process_steps)
        final_pypeitImage.build_mask(final_pypeitImage.image, final_pypeitImage.ivar,
                               saturation=nonlinear_counts, #self.spectrograph.detector[self.det-1]['saturation'],
                               mincounts=self.spectrograph.detector[self.det-1]['mincounts'])
        # Return
        return final_pypeitImage

    @property
    def nfiles(self):
        """
        Number of files in the files attribute

        Returns:
            int

        """
        return len(self.files) if isinstance(self.files, (np.ndarray, list)) else 0





"""
Module to run tests on PypeItPar classes
"""
import os
import numpy

import pytest

from pypeit.spectrographs.keck_deimos import KeckDEIMOSSpectrograph
from pypeit.tests.tstutils import dev_suite_required

@dev_suite_required
def test_deimosslitmask():
    f = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'keck_deimos', '830G_M_8500',
                     'DE.20100913.22358.fits.gz')
    spec = KeckDEIMOSSpectrograph()
    spec.get_slitmask(f)
    assert spec.slitmask.nslits == 106, 'Incorrect number of slits read!'




"""
Module to run tests on PypeItPar classes
"""
import os
import numpy

import pytest

from pypeit.spectrographs.opticalmodel import DetectorMap
from pypeit.spectrographs.keck_deimos import DEIMOSCameraDistortion, KeckDEIMOSSpectrograph

from pypeit.tests.tstutils import dev_suite_required


def test_detectormap():
    d = DetectorMap()
    xpix = numpy.array([[24,20],[24,20]])
    ypix = numpy.array([[1000,996],[1000,996]])
    ximg, yimg = d.image_coordinates(xpix, ypix)
    assert ximg.shape == yimg.shape, 'Did not return correct shape.'
    det, _xpix, _ypix = d.ccd_coordinates(ximg, yimg)
    assert numpy.all(numpy.isclose(_xpix, xpix) & numpy.isclose(_ypix, ypix)), 'I/O mismatch'


def test_deimos_distortion():
    c = DEIMOSCameraDistortion()
    assert abs(c.apply_distortion(c.remove_distortion(0.5)) - 0.5) < 1e-5, \
        'Failed distortion recovery'


@dev_suite_required
def test_deimos_mask_coordinates():
    f = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'keck_deimos', '830G_M_8500',
                     'DE.20100913.22358.fits.gz')
    spec = KeckDEIMOSSpectrograph()
    spec.get_grating(f)
    assert numpy.isclose(spec.grating.central_wave, 8500.0078125), 'Incorrect grating setup'
    # Get the chip and pixel coordinates (1-indexed!) at the central
    # wavelength
    ximg, yimg, ccd, xpix, ypix = spec.mask_to_pixel_coordinates(x=-127.15775, y=20.0565)
    assert ccd == 6, 'Incorrect chip selected'
    assert numpy.isclose(xpix, 528.54329982), 'Incorrect x coordinate'
    assert numpy.isclose(ypix, 487.64455754), 'Incorrect y coordinate'


@dev_suite_required
def test_deimos_ccd_slits():
    f = os.path.join(os.environ['PYPEIT_DEV'], 'RAW_DATA', 'keck_deimos', '830G_M_8500',
                     'DE.20100913.22358.fits.gz')
    spec = KeckDEIMOSSpectrograph()
    ximg, yimg, ccd, xpix, ypix = spec.mask_to_pixel_coordinates(filename=f)
    indx = ccd < 1
    assert numpy.sum(indx) == 5, 'Should find 4 slits that are off all CCDs at the central lambda'
    ccds_with_slits_at_cwl, n_per_ccd = numpy.unique(ccd, return_counts=True)
    assert len(set(numpy.arange(8)+1) - set(ccds_with_slits_at_cwl)) == 0, \
            'Should find slits on all CCDs'
    assert numpy.sum(n_per_ccd) == 106, 'Should return coordinates for all 106 slits'
    assert n_per_ccd[1] == 11, 'Incorrect number of slits on CCD 1'


